# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# python
import datetime
import json
import mock
import os
import re
import time

import requests
import pytest

# datadog
from datadog import initialize

from .constants import MONITOR_REFERENCED_IN_SLO_MESSAGE

WAIT_TIME = 10


class TestDatadog:
    host_name = "test.host.integration"

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def cleanup_roles(self, dog):
        """Prepare Azure Integration."""
        self.cleanup_role_uuids = []

        yield

        # Ensure we cleanup any resources we created during tests
        # These should be removed during tests, but here as well in case of test failures
        for uuid in self.cleanup_role_uuids:
            dog.Roles.delete(uuid)

    def test_tags(self, dog, get_with_retry, freezer):
        with freezer:
            hostname = "test.tags.host" + str(int(time.time()))

        # post a metric to make sure the test host context exists
        dog.Metric.send(metric="test.tag.metric", points=1, host=hostname)
        # Wait for host to appear
        get_with_retry("Tag", hostname)

        # Ready to test
        dog.Tag.create(hostname, tags=["test_tag:1", "test_tag:2"], source="datadog")
        get_with_retry(
            "Tag",
            hostname,
            retry_condition=lambda r: "test_tag:1" not in r["tags"]
            or "test_tag:2" not in r["tags"],
            retry_limit=30,
            source="datadog",
        )

        # The response from `update` can be flaky, so let's test that it work by getting the tags
        dog.Tag.update(hostname, tags=["test_tag:3"], source="datadog")
        get_with_retry(
            "Tag",
            hostname,
            retry_condition=lambda r: r["tags"] != ["test_tag:3"],
            retry_limit=30,
            source="datadog",
        )

        all_tags = dog.Tag.get_all()
        assert "tags" in all_tags

        assert (
            dog.Tag.delete(hostname, source="datadog") is None
        )  # Expect no response body on success

    @pytest.mark.vcr(match_on=("method", "scheme", "host", "port", "path", "query", "body"))
    def test_events(self, dog, get_with_retry, freezer):
        with freezer:
            now_ts = int(time.time())
            before_ts = now_ts - datetime.timedelta(minutes=5).total_seconds()

        now_title = "end test title " + str(now_ts)
        now_message = "test message " + str(now_ts)

        before_title = "start test title " + str(before_ts)
        before_message = "test message " + str(before_ts)

        now_event = dog.Event.create(
            title=now_title, text=now_message, date_happened=now_ts
        )
        before_event = dog.Event.create(
            title=before_title, text=before_message, date_happened=before_ts
        )

        assert now_event["event"]["title"] == now_title
        assert now_event["event"]["text"] == now_message
        assert now_event["event"]["date_happened"] == now_ts
        assert before_event["event"]["title"] == before_title
        assert before_event["event"]["text"] == before_message
        assert before_event["event"]["date_happened"] == before_ts

        # The returned event doesn"t contain host information, we need to get it separately
        event_id = dog.Event.create(
            title="test host", text="test host", host=self.host_name
        )["event"]["id"]
        event = get_with_retry("Event", event_id)
        assert event["event"]["host"] == self.host_name

        event_id = dog.Event.create(
            title="test no hostname",
            text="test no hostname",
            attach_host_name=False,
            alert_type="success",
        )["event"]["id"]
        event = get_with_retry("Event", event_id)
        assert not event["event"]["host"]
        assert event["event"]["alert_type"] == "success"

        event = dog.Event.create(
            title="test tags", text="test tags", tags=["test_tag:1", "test_tag:2"]
        )
        assert "test_tag:1" in event["event"]["tags"]
        assert "test_tag:2" in event["event"]["tags"]

        event = dog.Event.create(
            title="test source",
            text="test source",
            source_type_name="vsphere",
            priority="low",
        )
        event_id = event["event"]["id"]
        now_ts = event["event"]["date_happened"]
        get_with_retry("Event", event_id)
        events = dog.Event.query(
            start=now_ts - 100, end=now_ts + 100, priority="low", sources="vsphere"
        )
        assert events["events"], "No events found in stream"
        assert event_id in [event["id"] for event in events["events"]]

    def test_comments(self, dog, get_with_retry, freezer, user_handle):
        with freezer:
            now_ts = int(time.time())

        message = "test message " + str(now_ts)

        comment = dog.Comment.create(handle=user_handle, message=message)
        comment_id = comment["comment"]["id"]
        assert comment["comment"]["message"] == message

        get_with_retry("Event", comment_id)
        comment = dog.Comment.update(
            comment_id, handle=user_handle, message=message + " updated"
        )
        assert comment["comment"]["message"] == message + " updated"
        reply = dog.Comment.create(
            handle=user_handle, message=message + " reply", related_event_id=comment_id
        )
        assert reply["comment"]["message"] == message + " reply"

    def test_timeboard(self, dog, get_with_retry):
        graph = {
            "title": "test metric graph",
            "definition": {
                "requests": [{"q": "testing.metric.1{host:blah.host.1}"}],
                "viz": "timeseries",
            },
        }

        timeboard = dog.Timeboard.create(
            title="api timeboard", description="my api timeboard", graphs=[graph]
        )
        assert "api timeboard" == timeboard["dash"]["title"]
        assert "my api timeboard" == timeboard["dash"]["description"]
        assert timeboard["dash"]["graphs"][0] == graph

        timeboard = get_with_retry("Timeboard", timeboard["dash"]["id"])
        assert "api timeboard" == timeboard["dash"]["title"]
        assert "my api timeboard" == timeboard["dash"]["description"]
        assert timeboard["dash"]["graphs"][0] == graph

        graph = {
            "title": "updated test metric graph",
            "definition": {
                "requests": [{"q": "testing.metric.1{host:blah.host.1}"}],
                "viz": "timeseries",
            },
        }

        timeboard = dog.Timeboard.update(
            timeboard["dash"]["id"],
            title="updated api timeboard",
            description="my updated api timeboard",
            graphs=[graph],
        )

        assert "updated api timeboard" == timeboard["dash"]["title"]
        assert "my updated api timeboard" == timeboard["dash"]["description"]
        assert timeboard["dash"]["graphs"][0] == graph

        # Query all dashboards and make sure it"s in there.
        timeboards = dog.Timeboard.get_all()["dashes"]
        ids = [str(timeboard["id"]) for timeboard in timeboards]
        assert str(timeboard["dash"]["id"]) in ids

        assert dog.Timeboard.delete(timeboard["dash"]["id"]) is None

    def test_search(self, dog):
        results = dog.Infrastructure.search(q="")
        assert len(results["results"]["hosts"]) > 0
        assert len(results["results"]["metrics"]) > 0

    def test_metrics_simple(self, dog, get_with_retry, freezer):
        with freezer:
            now_ts = int(time.time())

            metric_name_single = "test.metric_single." + str(now_ts)
            host_name = "test.host." + str(now_ts)

            def retry_condition(r):
                return not r["series"]

            # Send metrics with single and multi points, and with compression
            assert (
                dog.Metric.send(metric=metric_name_single, points=1, host=host_name)[
                    "status"
                ]
                == "ok"
            )

        metric_query_single = get_with_retry(
            "Metric",
            operation="query",
            retry_condition=retry_condition,
            retry_limit=60,
            start=now_ts - 600,
            end=now_ts + 600,
            query="{}{{host:{}}}".format(metric_name_single, host_name),
        )
        assert len(metric_query_single["series"]) == 1
        assert metric_query_single["series"][0]["metric"] == metric_name_single
        assert metric_query_single["series"][0]["scope"] == "host:{}".format(host_name)
        assert len(metric_query_single["series"][0]["pointlist"]) == 1
        assert metric_query_single["series"][0]["pointlist"][0][1] == 1

    def test_metrics_list(self, dog, get_with_retry, freezer):
        with freezer:
            now_ts = int(time.time())

        metric_name_list = "test.metric_list." + str(now_ts)
        host_name = "test.host." + str(now_ts)

        def retry_condition(r):
            return not r["series"]

        points = [(now_ts - 60, 1), (now_ts, 2)]
        assert (
            dog.Metric.send(metric=metric_name_list, points=points, host=host_name)[
                "status"
            ]
            == "ok"
        )
        metric_query_list = get_with_retry(
            "Metric",
            operation="query",
            retry_condition=retry_condition,
            retry_limit=60,
            start=now_ts - 600,
            end=now_ts + 600,
            query="{}{{host:{}}}".format(metric_name_list, host_name),
        )
        assert len(metric_query_list["series"]) == 1
        assert metric_query_list["series"][0]["metric"] == metric_name_list
        assert metric_query_list["series"][0]["scope"] == "host:{}".format(host_name)
        assert len(metric_query_list["series"][0]["pointlist"]) == 2
        assert metric_query_list["series"][0]["pointlist"][0][1] == 1
        assert metric_query_list["series"][0]["pointlist"][1][1] == 2

    def test_metrics_tuple(self, dog, get_with_retry, freezer):
        with freezer:
            now_ts = int(time.time())

        metric_name_tuple = "test.metric_tuple." + str(now_ts)
        host_name = "test.host." + str(now_ts)

        def retry_condition(r):
            return not r["series"]

        points = (now_ts - 60, 1)
        assert (
            dog.Metric.send(
                metric=metric_name_tuple,
                points=points,
                host=host_name,
                compress_payload=False,
            )["status"]
            == "ok"
        )
        metric_query_tuple = get_with_retry(
            "Metric",
            operation="query",
            retry_condition=retry_condition,
            retry_limit=60,
            start=now_ts - 600,
            end=now_ts + 600,
            query="{}{{host:{}}}".format(metric_name_tuple, host_name),
        )
        assert len(metric_query_tuple["series"]) == 1
        assert metric_query_tuple["series"][0]["metric"] == metric_name_tuple
        assert metric_query_tuple["series"][0]["scope"] == "host:{}".format(host_name)
        assert len(metric_query_tuple["series"][0]["pointlist"]) == 1
        assert metric_query_tuple["series"][0]["pointlist"][0][1] == 1

    def test_distribution_metrics(self, dog, freezer):
        with freezer:
            now_ts = int(time.time())

        metric_name = "test.distribution_metric." + str(now_ts)
        host_name = "test.host." + str(now_ts)

        # Submit a distribution metric
        assert dog.Distribution.send(
            distributions=[{
                'metric': metric_name,
                'points': [(now_ts - 60, [1.0])],
                'type': 'distribution',
                'host': host_name,
            }]
        )["status"] == "ok"

        # FIXME: Query and verify the test metric result. Currently, it takes
        # too long for a new distribution metric to become available for query.

    def test_graph_snapshot(self, dog, get_with_retry, freezer):
        metric_query = "system.load.1{*}"
        event_query = "*"
        with freezer:
            end = int(time.time())
            start = end - 60 * 60  # go back 1 hour

        # Test without an event query
        snap = dog.Graph.create(metric_query=metric_query, start=start, end=end)
        assert "event_query" not in snap
        assert snap["metric_query"] == metric_query
        snapshot_url = snap["snapshot_url"]

        # Test with an event query
        snap = dog.Graph.create(
            metric_query=metric_query, start=start, end=end, event_query=event_query
        )
        assert snap["metric_query"] == metric_query
        assert snap["event_query"] == event_query
        snapshot_url = snap["snapshot_url"]

        # Test with a graph def
        graph_def = {
            "viz": "toplist",
            "requests": [
                {
                    "q": "top(system.disk.free{*} by {device}, 10, 'mean', 'desc')",
                    "style": {"palette": "dog_classic"},
                    "conditional_formats": [
                        {"palette": "red", "comparator": ">", "value": 50000000000},
                        {"palette": "green", "comparator": ">", "value": 30000000000},
                    ],
                }
            ],
        }
        graph_def = json.dumps(graph_def, sort_keys=True)
        snap = dog.Graph.create(graph_def=graph_def, start=start, end=end)
        assert "metric_query" not in snap
        assert "event_query" not in snap
        assert snap["graph_def"] == graph_def
        snapshot_url = snap["snapshot_url"]

        # Test snapshot status endpoint
        get_with_retry(
            "Graph",
            snapshot_url,
            operation="status",
            retry_condition=lambda r: r["status_code"] != 200,
            retry_limit=20,
        )

    def test_screenboard(self, vcr_cassette, dog, get_with_retry):
        def _compare_screenboard(apiBoard, expectedBoard):
            compare_keys = ["board_title", "height", "width", "widgets"]
            for key in compare_keys:
                assert apiBoard[key] == expectedBoard[key]

        board = {
            "width": 1024,
            "height": 768,
            "board_title": "datadog test",
            "widgets": [
                {
                    "type": "event_stream",
                    "title": False,
                    "height": 57,
                    "width": 59,
                    "y": 18,
                    "x": 84,
                    "query": "tags:release",
                    "time": {"live_span": "1w"},
                },
                {
                    "type": "image",
                    "height": 20,
                    "width": 32,
                    "y": 7,
                    "x": 32,
                    "url": "http://path/to/image.jpg",
                },
            ],
        }

        updated_board = {
            "width": 1024,
            "height": 768,
            "board_title": "datadog test",
            "widgets": [
                {
                    "type": "image",
                    "height": 20,
                    "width": 32,
                    "y": 7,
                    "x": 32,
                    "url": "http://path/to/image.jpg",
                }
            ],
        }

        create_res = dog.Screenboard.create(**board)
        _compare_screenboard(board, create_res)

        get_res = get_with_retry("Screenboard", create_res["id"])
        _compare_screenboard(get_res, create_res)
        assert get_res["id"] == create_res["id"]

        get_all_res = dog.Screenboard.get_all()["screenboards"]
        created = [s for s in get_all_res if s["id"] == create_res["id"]]
        assert len(created) == 1

        update_res = dog.Screenboard.update(get_res["id"], **updated_board)
        _compare_screenboard(update_res, updated_board)
        assert get_res["id"] == update_res["id"]

        share_res = dog.Screenboard.share(get_res["id"])
        assert share_res["board_id"] == get_res["id"]
        public_url = share_res["public_url"]

        if vcr_cassette.record_mode != "none":
            time.sleep(WAIT_TIME)
        response = requests.get(public_url)
        assert response.status_code == 200

        dog.Screenboard.revoke(get_res["id"])
        if vcr_cassette.record_mode != "none":
            time.sleep(WAIT_TIME)
        response = requests.get(public_url)
        assert response.status_code == 404

        delete_res = dog.Screenboard.delete(update_res["id"])
        assert delete_res["id"] == update_res["id"]

    def test_monitor_crud(self, dog, get_with_retry, freezer):
        # Metric alerts
        query = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 100"

        with freezer:
            options = {
                "silenced": {"*": int(time.time()) + 60 * 60},
                "notify_no_data": False,
            }
        monitor = dog.Monitor.create(type="metric alert", query=query, options=options)
        assert monitor["query"] == query
        assert monitor["options"]["notify_no_data"] == options["notify_no_data"]
        assert monitor["options"]["silenced"] == options["silenced"]

        monitor = get_with_retry("Monitor", monitor["id"])
        assert monitor["query"] == query
        assert monitor["options"]["notify_no_data"] == options["notify_no_data"]
        assert monitor["options"]["silenced"] == options["silenced"]

        query2 = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 200"
        monitor = dog.Monitor.update(monitor["id"], query=query2, options=options)
        assert monitor["query"] == query2
        assert monitor["options"]["notify_no_data"] == options["notify_no_data"]
        assert monitor["options"]["silenced"] == options["silenced"]

        name = "test_monitors"
        monitor = dog.Monitor.update(
            monitor["id"], query=query2, name=name, options={"notify_no_data": True}
        )
        assert monitor["name"] == name
        assert monitor["query"] == query2
        assert monitor["options"]["notify_no_data"] is True

        monitors = [m for m in dog.Monitor.get_all() if m["id"] == monitor["id"]]
        assert len(monitors) == 1

        assert dog.Monitor.delete(monitor["id"]) == {
            "deleted_monitor_id": monitor["id"]
        }

    def test_monitor_validate(self, dog, get_with_retry):
        monitor_type = "metric alert"
        valid_options = {"thresholds": {"critical": 200.0}}
        invalid_options = {"thresholds": {"critical": 90.0}}

        # Check with an invalid query.
        invalid_query = "THIS IS A BAD QUERY"
        res = dog.Monitor.validate(type=monitor_type, query=invalid_query, options=valid_options)
        assert res == {"errors": ["The value provided for parameter 'query' is invalid"]}

        # Check with a valid query, invalid options.
        valid_query = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 200"
        res = dog.Monitor.validate(type=monitor_type, query=valid_query, options=invalid_options)
        assert res == {"errors": ["Alert threshold (90.0) does not match that used in the query (200.0)."]}

        # Check with a valid query, valid options.
        res = dog.Monitor.validate(type=monitor_type, query=valid_query, options=valid_options)
        assert res == {}

    def test_monitor_can_delete(self, dog, freezer):
        # Create a monitor.
        query = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 100"
        with freezer:
            options = {
                "silenced": {"*": int(time.time()) + 60 * 60},
                "notify_no_data": False,
            }
        monitor = dog.Monitor.create(type="metric alert", query=query, options=options)

        # Check if you can delete the monitor.
        monitor_ids = [monitor["id"]]
        assert dog.Monitor.can_delete(monitor_ids=monitor_ids) == {
            "data": {"ok": monitor_ids},
            "errors": None,
        }

        # Create a monitor-based SLO.
        with freezer as dt:
            name = "test SLO {}".format(int(time.time()))

        thresholds = [{"timeframe": "7d", "target": 90}]
        slo = dog.ServiceLevelObjective.create(
            type="monitor",
            monitor_ids=monitor_ids,
            thresholds=thresholds,
            name=name,
        )["data"][0]

        # Check if you can delete the monitor.
        monitor_ids = [monitor["id"]]
        resp = dog.Monitor.can_delete(monitor_ids=monitor_ids)
        assert "errors" in resp
        assert str(monitor["id"]) in resp["errors"]
        assert len(resp["errors"][str(monitor["id"])])
        assert "is referenced in slos" in resp["errors"][str(monitor["id"])][0]

        # Delete the SLO.
        dog.ServiceLevelObjective.delete(slo["id"])

        # Check if you can delete the monitor.
        monitor_ids = [monitor["id"]]
        assert dog.Monitor.can_delete(monitor_ids=monitor_ids) == {
            "data": {"ok": monitor_ids},
            "errors": None,
        }

        # Delete the monitor to clean up the test.
        assert dog.Monitor.delete(monitor["id"]) == {
            "deleted_monitor_id": monitor["id"]
        }

    def test_monitor_can_delete_with_force(self, dog, freezer):
        # Create a monitor.
        query = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 100"
        with freezer:
            options = {
                "silenced": {"*": int(time.time()) + 60 * 60},
                "notify_no_data": False,
            }
        monitor = dog.Monitor.create(type="metric alert", query=query, options=options)
        monitor_ids = [monitor["id"]]

        # Create a monitor-based SLO.
        with freezer:
            name = "test SLO {}".format(int(time.time()))
        thresholds = [{"timeframe": "7d", "target": 90}]
        slo = dog.ServiceLevelObjective.create(
            type="monitor",
            monitor_ids=monitor_ids,
            thresholds=thresholds,
            name=name,
        )["data"][0]

        # Check if you can delete the monitor with force option
        assert dog.Monitor.delete(monitor["id"], force=True) == {
            "deleted_monitor_id": monitor["id"]
        }

        # Delete the SLO.
        dog.ServiceLevelObjective.delete(slo["id"])

    def test_service_level_objective_crud(self, dog, freezer):
        numerator = "sum:my.custom.metric{type:good}.as_count()"
        denominator = "sum:my.custom.metric{*}.as_count()"
        query = {"numerator": numerator, "denominator": denominator}
        thresholds = [{"timeframe": "7d", "target": 90}]
        with freezer:
            name = "test SLO {}".format(int(time.time()))
        slo = dog.ServiceLevelObjective.create(
            type="metric",
            query=query,
            thresholds=thresholds,
            name=name,
            tags=["type:test"],
        )["data"][0]
        assert slo["name"] == name

        numerator2 = "sum:my.custom.metric{type:good,!type:ignored}.as_count()"
        denominator2 = "sum:my.custom.metric{!type:ignored}.as_count()"
        query = {"numerator": numerator2, "denominator": denominator2}
        slo = dog.ServiceLevelObjective.update(
            id=slo["id"],
            type="metric",
            query=query,
            thresholds=thresholds,
            name=name,
            tags=["type:test"],
        )["data"][0]
        assert slo["name"] == name
        slos = [
            s
            for s in dog.ServiceLevelObjective.get_all()["data"]
            if s["id"] == slo["id"]
        ]
        assert len(slos) == 1

        assert dog.ServiceLevelObjective.get(slo["id"])["data"]["id"] == slo["id"]
        dog.ServiceLevelObjective.delete(slo["id"])

    @pytest.mark.admin_needed
    def test_monitor_muting(self, dog, get_with_retry):
        query1 = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 100"
        query2 = "avg(last_1h):sum:system.net.bytes_rcvd{*} by {host} > 100"
        monitor1 = dog.Monitor.create(type="metric alert", query=query1)
        monitor2 = dog.Monitor.create(type="metric alert", query=query2)

        dt = dog.Monitor.mute_all()
        assert dt["active"] is True
        assert dt["scope"] == ["*"]

        assert dog.Monitor.unmute_all() is None  # No response expected

        monitor1 = dog.Monitor.mute(monitor1["id"])
        assert monitor1["options"]["silenced"] == {"*": None}

        monitor2 = dog.Monitor.mute(monitor2["id"], scope="host:foo")
        assert monitor2["options"]["silenced"] == {"host:foo": None}

        get_with_retry(
            "Monitor",
            monitor2["id"],
            retry_condition=lambda r: r["options"]["silenced"] != {"host:foo": None},
        )
        monitor2 = dog.Monitor.unmute(monitor2["id"], scope="host:foo")
        assert monitor2["options"]["silenced"] == {}

        dog.Monitor.delete(monitor1["id"])
        dog.Monitor.delete(monitor2["id"])

    def test_downtime(self, dog, get_with_retry, freezer):
        with freezer:
            start = int(time.time())
            end = start + 1000

        # Create downtime
        downtime = dog.Downtime.create(scope="test_tag:1", start=start, end=end)
        assert downtime["start"] == start
        assert downtime["end"] == end
        assert downtime["scope"] == ["test_tag:1"]
        assert downtime["disabled"] is False

        get_with_retry("Downtime", downtime["id"])

        # Update downtime
        message = "Doing some testing on staging."
        with freezer as dt:
            dt.tick()
            end = int(time.time()) + 60000

        downtime = dog.Downtime.update(
            downtime["id"], scope="test_tag:2", end=end, message=message
        )
        assert downtime["end"] == end
        assert downtime["message"] == message
        assert downtime["scope"] == ["test_tag:2"]
        assert downtime["disabled"] is False

        # Delete downtime
        assert dog.Downtime.delete(downtime["id"]) is None
        downtime = get_with_retry(
            "Downtime", downtime["id"], retry_condition=lambda r: r["disabled"] is False
        )

    def test_downtime_cancel_by_scope(self, dog, get_with_retry, freezer):
        scope_one = "test:integration_one"
        scope_two = "test:integration_two"
        with freezer:
            start = int(time.time())

        # Create downtime with scope_one
        end = start + 1000
        downtime_one = dog.Downtime.create(scope=scope_one, start=start, end=end)
        assert downtime_one["scope"] == [scope_one]
        assert downtime_one["disabled"] is False

        # Create downtime with scope_one
        with freezer as dt:
            dt.tick()
            end = int(time.time()) + 60000
        downtime_two = dog.Downtime.create(scope=scope_one, start=start, end=end)
        assert downtime_two["scope"] == [scope_one]
        assert downtime_two["disabled"] is False

        with freezer as dt:
            dt.tick()
            end = int(time.time()) + 120000
        downtime_three = dog.Downtime.create(scope=scope_two, start=start, end=end)
        assert downtime_three["scope"] == [scope_two]
        assert downtime_three["disabled"] is False

        downtimes_with_scope_one = [downtime_one, downtime_two]
        downtimes_with_scope_two = [downtime_three]

        # Cancel downtimes with scope `scope_one`
        dog.Downtime.cancel_downtime_by_scope(scope=scope_one)

        # Verify only the downtimes with scope `scope_one` were canceled
        for downtime in downtimes_with_scope_one:
            get_with_retry(
                "Downtime", downtime["id"], retry_condition=lambda r: r["disabled"] is False
            )
        for downtime in downtimes_with_scope_two:
            d = get_with_retry("Downtime", downtime["id"])
            assert d["disabled"] is False

        # Cancel downtimes with scope `scope_two`
        dog.Downtime.cancel_downtime_by_scope(scope=scope_two)

        # Verify downtimes with scope `scope_two` were canceled
        for downtime in downtimes_with_scope_two:
            get_with_retry(
                "Downtime", downtime["id"], retry_condition=lambda r: r["disabled"] is False
            )

    def test_service_check(self, dog):
        assert dog.ServiceCheck.check(
            check="check_pg",
            host_name="host0",
            status=1,
            message="PG is WARNING",
            tags=["db:prod_data"],
        ) == {"status": "ok"}

    def test_host_muting(self, dog, get_with_retry, freezer):
        with freezer:
            end = int(time.time()) + 60 * 60
        hostname = "my.test.host" + str(end)
        message = "Muting this host for a test."

        # post a metric to make sure the test host context exists
        dog.Metric.send(metric="test.muting.host", points=1, host=hostname)
        # Wait for host to appear
        get_with_retry("Tag", hostname)

        # Mute a host
        mute = dog.Host.mute(hostname, end=end, message=message)
        assert mute["hostname"] == hostname
        assert mute["action"] == "Muted"
        assert mute["message"] == message
        assert mute["end"] == end

        # We shouldn"t be able to mute a host that"s already muted, unless we include
        # the override param.
        end2 = end + 60 * 15
        get_with_retry(
            "Host",
            hostname,
            operation="mute",
            retry_condition=lambda r: "errors" not in r,
            end=end2,
        )
        mute = dog.Host.mute(hostname, end=end2, override=True)
        assert mute["hostname"] == hostname
        assert mute["action"] == "Muted"
        assert mute["end"] == end2

        unmute = dog.Host.unmute(hostname)
        assert unmute["hostname"] == hostname
        assert unmute["action"] == "Unmuted"

    def test_get_all_embeds(self, dog):
        all_embeds = dog.Embed.get_all()
        # Check all embeds is a valid response
        assert "embedded_graphs" in all_embeds

    def test_embed_crud(self, dog, get_with_retry):
        # Initialize a graph definition
        graph_def = {
            "viz": "toplist",
            "requests": [
                {
                    "q": "top(system.disk.free{$var} by {device}, 10, 'mean', 'desc')",
                    "style": {"palette": "dog_classic"},
                    "conditional_formats": [
                        {"palette": "red", "comparator": ">", "value": 50000000000},
                        {"palette": "green", "comparator": ">", "value": 30000000000},
                    ],
                }
            ],
        }
        timeframe = "1_hour"
        size = "medium"
        legend = "no"
        title = "Custom titles!"
        # Dump the dictionary to a JSON string and make an API call
        graph_json = json.dumps(graph_def, sort_keys=True)
        embed = dog.Embed.create(
            graph_json=graph_json,
            timeframe=timeframe,
            size=size,
            legend=legend,
            title=title,
        )
        # Check various embed attributes
        assert "embed_id" in embed
        assert embed["revoked"] is False
        assert len(embed["template_variables"]) == 1
        assert embed["template_variables"][0] == "var"
        assert "html" in embed
        assert embed["graph_title"] == title

        var = "asdfasdfasdf"
        response_graph = get_with_retry("Embed", embed["embed_id"], var=var)
        # Check the graph has the same embed_id and the template_var is added to the url
        assert "embed_id" in response_graph
        assert response_graph["embed_id"] == embed["embed_id"]
        assert len(response_graph["html"]) > len(embed["html"])
        assert var in response_graph["html"]

        assert "success" in dog.Embed.enable(embed["embed_id"])

        assert "success" in dog.Embed.revoke(embed["embed_id"])

    @pytest.mark.admin_needed
    def test_user_crud(self, dog, get_with_retry, freezer):
        with freezer:
            now = int(time.time())
        handle = "user{}@test.com".format(now)
        name = "Test User"
        alternate_name = "Test User Alt"

        # test create user
        user = dog.User.create(handle=handle, name=name, access_role="ro")
        assert "user" in user
        assert user["user"]["handle"] == handle
        assert user["user"]["name"] == name
        assert user["user"]["disabled"] is False
        assert user["user"]["access_role"] == "ro"

        # test get user
        user = get_with_retry("User", handle)
        assert "user" in user
        assert user["user"]["handle"] == handle
        assert user["user"]["name"] == name

        # test update user
        user = dog.User.update(handle, name=alternate_name, access_role="st")
        assert user["user"]["handle"] == handle
        assert user["user"]["name"] == alternate_name
        assert user["user"]["disabled"] is False
        assert user["user"]["access_role"] == "st"

        # test disable user
        dog.User.delete(handle)
        u = dog.User.get(handle)
        assert "user" in u
        assert u["user"]["disabled"] is True

        # test get all users
        u = dog.User.get_all()
        assert "users" in u
        assert len(u["users"]) >= 1

    @pytest.mark.admin_needed
    def test_roles_crud(self, dog):
        role_name = "test_role"

        data = {
                "type": "roles",
                "attributes": {
                    "name": role_name
                }
            }

        # test create role
        role = dog.Roles.create(data=data)
        self.cleanup_role_uuids.append(role["data"]["id"])
        assert "roles" in role["data"]["type"]
        assert role["data"]["id"] is not None
        assert role["data"]["attributes"]["name"] == role_name

        role_uuid = role["data"]["id"]

        # test update role
        new_role_name = "test_role_2"
        data = {
                "type": "roles",
                "attributes": {
                    "name": new_role_name,
                    "id": role_uuid,
                }
            }

        role = dog.Roles.update(role_uuid, data=data)
        assert "roles" in role["data"]["type"]
        assert role["data"]["id"] is not None
        assert role["data"]["attributes"]["name"] == new_role_name

        # test assign permission

        permissions = dog.Permissions.get_all()
        assert "permissions" in permissions["data"][0]["type"]
        assert len(permissions["data"]) > 0

        permission_uuid = permissions["data"][0]["id"]
        data = {
                    "type": "permissions",
                    "id": permission_uuid
                }

        role = dog.Roles.assign_permission(role_uuid, data=data)
        assert "permissions" in role["data"][0]["type"]

        # test unassign permission
        data = {
            "type": "permissions",
            "id": permission_uuid
        }

        role = dog.Roles.unassign_permission(role_uuid, data=data)
        assert "permissions" in role["data"][0]["type"]
        assert len(permissions["data"]) > 0


        # test delete role
        dog.Roles.delete(role_uuid)

        # check if new role is deleted successfully
        res = dog.Roles.get(role_uuid)
        assert "errors" in res

    @mock.patch('datadog.api._return_raw_response', True)
    def test_user_agent(self, dog):
        _, resp = dog.api_client.APIClient.submit('GET', 'validate')
        assert re.match(r'^datadogpy\/[^\s]+ \(python [^\s]+; os [^\s]+; arch [^\s]+\)$', resp.request.headers['User-Agent'])



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import time
# import unittest
import threading

from datadog import lambda_metric, datadog_lambda_wrapper
from datadog.threadstats.aws_lambda import _lambda_stats


TOTAL_NUMBER_OF_THREADS = 1000


class MemoryReporter(object):
    """ A reporting class that reports to memory for testing. """

    def __init__(self):
        self.distributions = []
        self.dist_flush_counter = 0

    def flush_distributions(self, dists):
        self.distributions += dists
        self.dist_flush_counter = self.dist_flush_counter + 1


@datadog_lambda_wrapper
def wrapped_function(id):
    lambda_metric("dist_" + str(id), 42)
    # sleep makes the os continue another thread
    time.sleep(0.001)

    lambda_metric("common_dist", 42)


# Lambda wrapper - mute thread safety test, python 2.7 issues
# class TestWrapperThreadSafety(unittest.TestCase):

#     def test_wrapper_thread_safety(self):
#         _lambda_stats.reporter = MemoryReporter()

#         for i in range(TOTAL_NUMBER_OF_THREADS):
#             threading.Thread(target=wrapped_function, args=[i]).start()
#         # Wait all threads to finish
#         time.sleep(10)

#         # Check that at least one flush happened
#         self.assertGreater(_lambda_stats.reporter.dist_flush_counter, 0)

#         dists = _lambda_stats.reporter.distributions
#         self.assertEqual(len(dists), TOTAL_NUMBER_OF_THREADS + 1)



# coding: utf8
# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import logging
import unittest

from mock import patch

from datadog.util.compat import conditional_lru_cache, is_higher_py32

class TestConditionalLRUCache(unittest.TestCase):
    def test_normal_usage(self):
        @conditional_lru_cache
        def test_function(some_string, num1, num2, num3):
            return (some_string, num1 + num2 + num3)

        for idx in range(600):
            self.assertEqual(
                test_function("abc", idx, idx*2, idx *3),
                ("abc", idx + idx * 2 + idx *3),
            )

    def test_var_args(self):
        @conditional_lru_cache
        def test_function(*args):
            return sum(list(args))

        args = []
        for idx in range(100):
            args.append(idx)
            self.assertEqual(
                test_function(*args),
                sum(args),
            )

    # pylint: disable=no-self-use
    def test_debug_log(self):
        test_object_logger = logging.getLogger('datadog.util')
        with patch.object(test_object_logger, 'debug') as mock_debug:
            @conditional_lru_cache
            def test_function():
                pass

            test_function()

            if is_higher_py32():
                mock_debug.assert_called_once()
            else:
                mock_debug.assert_not_called()



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
Helper(s), load fixtures.
"""
# stdlib
import os


def load_fixtures(name):
    """
    Load fixtures.

    Args:
        name (string): name of the fixture
    """
    with open(os.path.join(os.path.dirname(__file__), 'fixtures', '{}'.format(name))) as fixture:
        return fixture.read()



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import os

import pytest


class TestSynthetics:

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def synthetics(self, dog):
        # config and options for an API test
        self.options = {"tick_every": 300}
        self.config = {
            "assertions": [{"operator": "is", "type": "statusCode", "target": 200}],
            "request": {"method": "GET", "url": "https://example.com", "timeout": 30},
        }

        # config and option for a Browser test
        self.options_browser = {"device_ids": ["laptop_large"], "tick_every": 900}

        # create an API test
        self.output = dog.Synthetics.create_test(
            config=self.config,
            locations=["aws:us-east-2"],
            message="Test API",
            options=self.options,
            tags=["test:synthetics_api"],
            type="api",
            name="Test with API",
        )

        # create a Browser test
        self.output_browser = dog.Synthetics.create_test(
            config=self.config,
            locations=["aws:us-east-2"],
            message="Test Browser",
            options=self.options_browser,
            tags=["test:synthetics_browser"],
            type="browser",
            name="Test with Browser",
        )

        self.public_test_id = self.output["public_id"]
        self.public_test_id_browser = self.output_browser["public_id"]

        yield

        # delete all tests present in the account if any
        self.output_cleanup = dog.Synthetics.get_all_tests()
        self.public_ids_test_to_delete = []
        for test in self.output_cleanup["tests"]:
            self.public_ids_test_to_delete.append(test["public_id"])
        dog.Synthetics.delete_test(public_ids=self.public_ids_test_to_delete)

    def test_get_update_pause_test(self, dog):
        # test that both tests are live
        assert len(self.output) > 1
        assert "public_id" in self.output
        assert self.output["status"] == "live"
        assert len(self.output_browser) > 1
        assert "public_id" in self.output_browser
        assert self.output_browser["status"] == "paused"

        # get this newly created tests
        output_api = dog.Synthetics.get_test(id=self.public_test_id)
        assert "public_id" in output_api
        assert output_api["status"] == "live"
        output_browser = dog.Synthetics.get_test(id=self.public_test_id_browser)
        assert "public_id" in output_browser
        assert output_browser["status"] == "paused"

        # test that we can retrieve results_ids
        output_api = dog.Synthetics.get_results(id=self.public_test_id)
        assert output_api["results"] is not None
        output_browser = dog.Synthetics.get_results(id=self.public_test_id_browser)
        assert output_browser["results"] is not None

        # edit the API test
        self.options = {"tick_every": 60}
        self.config["assertions"] = [
            {"operator": "isNot", "type": "statusCode", "target": 404}
        ]

        output = dog.Synthetics.edit_test(
            id=self.public_test_id,
            config=self.config,
            type="api",
            locations=["aws:us-west-2"],
            message="Test API edited",
            name="Test with API edited",
            options=self.options,
            tags=["test:edited"],
        )
        assert "error" not in output
        # test that the new name matches
        assert output["name"] == "Test with API edited"

        # edit the Browser test
        self.config["assertions"] = [
            {"operator": "isNot", "type": "statusCode", "target": 404}
        ]
        self.options_browser = {"device_ids": ["tablet"], "tick_every": 1800}

        output = dog.Synthetics.edit_test(
            id=self.public_test_id,
            config=self.config,
            type="api",
            locations=["aws:us-west-2"],
            message="Test Browser edited",
            name="Test Browser edited",
            options=self.options_browser,
            tags=["test:edited"],
        )
        assert "error" not in output
        # test that the new name matches
        assert output["name"] == "Test Browser edited"

        # pause the API test
        output = dog.Synthetics.start_or_pause_test(id=self.public_test_id, new_status="paused")
        # output is a boolean
        assert output == True

    def test_get_all_tests(self, dog):
        output = dog.Synthetics.get_all_tests()
        # 2 tests were created
        assert len(output["tests"]) >= 2

    def test_get_locations(self, dog):
        output = dog.Synthetics.get_locations()
        assert len(output) == 1
        # 13 regions
        assert len(output["locations"]) >= 10

    def test_get_devices(self, dog):
        output = dog.Synthetics.get_devices()
        assert len(output) == 1
        # 3 devices
        assert len(output["devices"]) >= 3

    def test_delete_test(self, dog):
        # delete the test
        output = dog.Synthetics.delete_test(public_ids=[self.public_test_id])
        assert output["deleted_tests"] is not None



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import re
import time
import threading

from datadog import ThreadStats


class MemoryReporter(object):
    """ A reporting class that reports to memory for testing. """

    def __init__(self):
        self.metrics = []
        self.events = []

    def flush_metrics(self, metrics):
        self.metrics += metrics

    def flush_events(self, events):
        self.events += events


class ThreadStatsTest(ThreadStats):
    def send_metrics_and_event(self, id):
        # Counter
        self.increment("counter", timestamp=12345)
        time.sleep(0.001)  # sleep makes the os continue another thread

        # Gauge
        self.gauge("gauge_" + str(id), 42)
        time.sleep(0.001)  # sleep makes the os continue another thread

        # Histogram
        self.histogram("histogram", id, timestamp=12345)
        time.sleep(0.001)  # sleep makes the os continue another thread

        # Event
        self.event("title", "content")


class TestThreadStatsThreadSafety(object):

    def test_threadstats_thread_safety(self):
        stats = ThreadStatsTest()
        stats.start(roll_up_interval=10, flush_in_thread=False)
        reporter = stats.reporter = MemoryReporter()

        for i in range(10000):
            threading.Thread(target=stats.send_metrics_and_event, args=[i]).start()
        # Wait all threads to finish
        time.sleep(10)

        # Flush and check
        stats.flush()
        metrics = reporter.metrics
        events = reporter.events

        # Overview
        assert len(metrics) == 10009

        # Sort metrics
        counter_metrics = []
        gauge_metrics = []
        histogram_metrics = []

        for m in metrics:
            if re.match("gauge_.*", m['metric']):
                gauge_metrics.append(m)
            elif re.match("histogram.*", m['metric']):
                histogram_metrics.append(m)
            else:
                counter_metrics.append(m)

        # Counter
        assert len(counter_metrics) == 1
        counter = counter_metrics[0]
        assert counter['points'][0][1] == 10000

        # Gauge
        assert len(gauge_metrics) == 10000

        # Histogram
        assert len(histogram_metrics) == 8
        count_histogram = filter(lambda x: x['metric'] == "histogram.count", histogram_metrics)[0]
        assert count_histogram['points'][0][1] == 10000
        sum_histogram = filter(lambda x: x['metric'] == "histogram.avg", histogram_metrics)[0]
        assert sum_histogram['points'][0][1] == 4999.5

        # Events
        assert 10000 == len(events)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import pytest

TEST_ACCOUNT_ID = "123456789101"
TEST_ROLE_NAME = "DatadogApiTestRole"
TEST_LAMBDA_ARN = "arn:aws:lambda:us-east-1:123456789101:function:APITest"
AVAILABLE_SERVICES = 6


class TestAwsLogsIntegration:

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def aws_integration(self, dog):
        """Prepare AWS Integration."""
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID,
            role_name=TEST_ROLE_NAME
        )
        yield
        dog.AwsIntegration.delete(account_id=TEST_ACCOUNT_ID, role_name=TEST_ROLE_NAME)

    def test_list_log_services(self, dog):
        output = dog.AwsLogsIntegration.list_log_services()
        assert len(output) >= AVAILABLE_SERVICES

    def test_aws_logs_crud(self, dog):
        add_lambda_arn_output = dog.AwsLogsIntegration.add_log_lambda_arn(
            account_id=TEST_ACCOUNT_ID,
            lambda_arn=TEST_LAMBDA_ARN
        )
        assert add_lambda_arn_output == {}
        save_services_output = dog.AwsLogsIntegration.save_services(
            account_id=TEST_ACCOUNT_ID,
            services=["s3", "elb", "elbv2", "cloudfront", "redshift", "lambda"]
        )
        assert save_services_output == {}
        list_output = dog.AwsLogsIntegration.list()
        expected_fields = [
            'services',
            'lambdas',
            'account_id'
        ]
        assert all(k in list_output[0].keys() for k in expected_fields)
        delete_output = dog.AwsLogsIntegration.delete_config(
            account_id=TEST_ACCOUNT_ID,
            lambda_arn=TEST_LAMBDA_ARN
        )
        assert delete_output == {}

    def test_check_lambda(self, dog):
        output = dog.AwsLogsIntegration.check_lambda(
            account_id=TEST_ACCOUNT_ID,
            lambda_arn=TEST_LAMBDA_ARN
        )
        assert 'status' in output.keys()

    def test_check_services(self, dog):
        output = dog.AwsLogsIntegration.check_services(
            account_id=TEST_ACCOUNT_ID,
            services=["s3", "elb", "elbv2", "cloudfront", "redshift", "lambda"]
        )
        assert 'status' in output.keys()



# Unless explicitly stated otherwise all files in this repository are licensed
# under the BSD-3-Clause License. This product includes software developed at
# Datadog (https://www.datadoghq.com/).

# Copyright 2021-Present Datadog, Inc

from threading import Event, Thread
import os
import time

# pylint: disable=import-error
import psutil


# pylint: disable=useless-object-inheritance
class SysInfoObserver(object):
    """
    SysInfoObserver collects timed CPU and memory usage stats in a separate
    thread about the current process.
    """

    def __init__(self, interval=0.1):
        self._stats = []

        self.interval = interval

        self.exit = None
        self.initial_cpu_user = None
        self.initial_cpu_system = None
        self.observer_thread = None
        self.proc_info = None

    def __enter__(self):
        if self.observer_thread:
            raise RuntimeError("Observer already running")

        self.exit = Event()

        pid = os.getpid()
        self.proc_info = psutil.Process(pid)

        # Record baselines
        self.initial_cpu_user = self.proc_info.cpu_times().user
        self.initial_cpu_system = self.proc_info.cpu_times().system
        self.initial_mem_rss = self.proc_info.memory_full_info().rss
        self.initial_mem_vms = self.proc_info.memory_full_info().vms

        self.observer_thread = Thread(
            name=self.__class__.__name__,
            target=self.poll_system_info,
            args=(
                self.proc_info,
                self.interval,
            ),
        )
        self.observer_thread.daemon = True
        self.observer_thread.start()

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.exit.set()
        self.observer_thread.join()

    def poll_system_info(self, proc_info, interval):
        while not self.exit.is_set():
            time.sleep(interval)

            mem_info = proc_info.memory_full_info()
            datapoint = {
                "interval": interval,
                "mem.rss_diff_kb": (mem_info.rss - self.initial_mem_rss) / 1024,
                "mem.vms_diff_kb": (mem_info.vms- self.initial_mem_vms)  / 1024,
            }

            self._stats.append(datapoint)

    @property
    def stats(self):
        # CPU data is cumulative
        agg_stats = {
            "cpu.user": self.proc_info.cpu_times().user - self.initial_cpu_user,
            "cpu.system": self.proc_info.cpu_times().system - self.initial_cpu_system,
        }

        if not self.exit.is_set():
            raise RuntimeError(
                "You can only collect aggregated stats after context manager exits"
            )

        datapoints = len(self._stats)
        for datapoint in self._stats:
            for key, val in datapoint.items():
                if key.startswith("cpu"):
                    continue

                if key not in agg_stats:
                    agg_stats[key] = 0.0

                agg_stats[key] += val

        for key, val in agg_stats.items():
            if not key.startswith("cpu"):
                agg_stats[key] = val / datapoints

        return agg_stats



# coding: utf8

# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import unittest
import mock
import os
import tempfile

from datadog.dogshell.wrap import OutputReader, build_event_body, parse_options, execute, Timeout, poll_proc
from datadog.util.compat import is_p3k


HERE = os.path.dirname(os.path.abspath(__file__))


class TestDogwrap(unittest.TestCase):
    def test_output_reader(self):
        with open(os.path.join(HERE, "fixtures", "proc_out.txt"), 'rb') as cmd_out:
            content = cmd_out.read()

        with tempfile.TemporaryFile() as fwd_out:
            reader = OutputReader(open(os.path.join(HERE, "fixtures", "proc_out.txt"), 'rb'), fwd_out)
            reader.start()
            reader.join()
            self.assertIsInstance(reader.content, bytes)
            self.assertEqual(reader.content, content)
            fwd_out.seek(0, 0)
            self.assertEqual(reader.content, fwd_out.read())

    def test_build_event_body(self):
        # Only cmd is already unicode, the rest is decoded in the function
        cmd = u"yö dudes"
        returncode = 0
        stdout = b"s\xc3\xb9p\xaa"
        stderr = b"d\xc3\xa0wg\xaa"
        notifications = b"@m\xc3\xa9\xaa"
        expected_body = u"%%%\n" \
            u"**>>>> CMD <<<<**\n```\nyö dudes \n```\n" \
            u"**>>>> EXIT CODE <<<<**\n\n 0\n\n\n" \
            u"**>>>> STDOUT <<<<**\n```\nsùp\ufffd \n```\n" \
            u"**>>>> STDERR <<<<**\n```\ndàwg\ufffd \n```\n" \
            u"**>>>> NOTIFICATIONS <<<<**\n\n @mé\ufffd\n" \
            u"%%%\n"

        event_body = build_event_body(cmd, returncode, stdout, stderr, notifications)
        self.assertEqual(expected_body, event_body)

        # notifications can be unicode already in py3, make sure we don't try decoding
        notifications = notifications.decode("utf-8", "replace")
        event_body = build_event_body(cmd, returncode, stdout, stderr, notifications)
        self.assertEqual(expected_body, event_body)

    def test_parse_options(self):
        options, cmd = parse_options([])
        self.assertEqual(cmd, '')

        # The output of parse_args is already unicode in python 3, so don't encode the input
        if is_p3k():
            arg = u'helløøééé'
        else:
            arg = u'helløøééé'.encode('utf-8')

        options, cmd = parse_options(['-n', 'name', '-k', 'key', '-m', 'all', '-p', 'low', '-t', '123',
                                      '--sigterm_timeout', '456', '--sigkill_timeout', '789',
                                      '--proc_poll_interval', '1.5', '--notify_success', 'success',
                                      '--notify_error', 'error', '-b', '--tags', 'k1:v1,k2:v2',
                                      'echo', arg])
        self.assertEqual(cmd, u'echo helløøééé')
        self.assertEqual(options.name, 'name')
        self.assertEqual(options.api_key, 'key')
        self.assertEqual(options.submit_mode, 'all')
        self.assertEqual(options.priority, 'low')
        self.assertEqual(options.timeout, 123)
        self.assertEqual(options.sigterm_timeout, 456)
        self.assertEqual(options.sigkill_timeout, 789)
        self.assertEqual(options.proc_poll_interval, 1.5)
        self.assertEqual(options.notify_success, 'success')
        self.assertEqual(options.notify_error, 'error')
        self.assertTrue(options.buffer_outs)
        self.assertEqual(options.tags, 'k1:v1,k2:v2')

        with self.assertRaises(SystemExit):
            parse_options(['-m', 'invalid'])

        with self.assertRaises(SystemExit):
            parse_options(['-p', 'invalid'])

        with self.assertRaises(SystemExit):
            parse_options(['-t', 'invalid'])

        with self.assertRaises(SystemExit):
            parse_options(['--sigterm_timeout', 'invalid'])

        with self.assertRaises(SystemExit):
            parse_options(['--sigkill_timeout', 'invalid'])

        with self.assertRaises(SystemExit):
            parse_options(['--proc_poll_interval', 'invalid'])

        with mock.patch.dict(os.environ, values={"DD_API_KEY": "the_key"}, clear=True):
            options, _ = parse_options([])
            self.assertEqual(options.api_key, "the_key")

    def test_poll_proc(self):
        mock_proc = mock.Mock()
        mock_proc.poll.side_effect = [None, 0]

        return_value = poll_proc(mock_proc, 0.1, 1)
        self.assertEqual(return_value, 0)
        self.assertEqual(mock_proc.poll.call_count, 2)

    def test_poll_timeout(self):
        mock_proc = mock.Mock()
        mock_proc.poll.side_effect = [None, None, None]

        with self.assertRaises(Timeout):
            poll_proc(mock_proc, 0.1, 0.2)

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.return_value = 0

        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, 0)
        self.assertEqual(stdout, b'out1\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_called_once_with(mock_proc, 1, 10)
        mock_proc.terminate.assert_not_called()
        mock_proc.kill.assert_not_called()

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute_exit_code(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'out2\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.return_value = 14

        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, 14)
        self.assertEqual(stdout, b'out1\nout2\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_called_once_with(mock_proc, 1, 10)
        mock_proc.terminate.assert_not_called()
        mock_proc.kill.assert_not_called()

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute_cmd_timeout(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'out2\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.side_effect = [Timeout, 1]

        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, Timeout)
        self.assertEqual(stdout, b'out1\nout2\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_has_calls([
            mock.call(mock_proc, 1, 10),
            mock.call(mock_proc, 1, 20)
        ])
        mock_proc.terminate.assert_called_once()
        mock_proc.kill.assert_not_called()

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute_sigterm_timeout(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'out2\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.side_effect = [Timeout, Timeout, 2]

        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, Timeout)
        self.assertEqual(stdout, b'out1\nout2\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_has_calls([
            mock.call(mock_proc, 1, 10),
            mock.call(mock_proc, 1, 20),
            mock.call(mock_proc, 1, 30)
        ])
        mock_proc.terminate.assert_called_once()
        mock_proc.kill.assert_called_once()

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute_sigkill_timeout(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'out2\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.side_effect = [Timeout, Timeout, Timeout]

        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, Timeout)
        self.assertEqual(stdout, b'out1\nout2\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_has_calls([
            mock.call(mock_proc, 1, 10),
            mock.call(mock_proc, 1, 20),
            mock.call(mock_proc, 1, 30)
        ])
        mock_proc.terminate.assert_called_once()
        mock_proc.kill.assert_called_once()

    @mock.patch('datadog.dogshell.wrap.poll_proc')
    @mock.patch('subprocess.Popen')
    def test_execute_oserror(self, mock_popen, mock_poll):
        mock_proc = mock.Mock()
        mock_proc.stdout.readline.side_effect = [b'out1\n', b'out2\n', b'']
        mock_proc.stderr.readline.side_effect = [b'err1\n', b'']
        mock_popen.return_value = mock_proc
        mock_poll.side_effect = [Timeout, Timeout]
        mock_proc.kill.side_effect = OSError(3, 'No process')
        return_code, stdout, stderr, duration = execute('foo', 10, 20, 30, 1, False)
        self.assertEqual(return_code, Timeout)
        self.assertEqual(stdout, b'out1\nout2\n')
        self.assertEqual(stderr, b'err1\n')

        mock_popen.assert_called_once()
        mock_poll.assert_has_calls([
            mock.call(mock_proc, 1, 10),
            mock.call(mock_proc, 1, 20)
        ])
        mock_proc.terminate.assert_called_once()
        mock_proc.kill.assert_called_once()

    @mock.patch('subprocess.Popen')
    def test_execute_popen_fail(self, mock_popen):
        mock_popen.side_effect = ValueError('Bad things')

        with self.assertRaises(ValueError):
            execute('sleep 1', 10, 1, 1, 1, False)



#!/usr/bin/env python
# Unless explicitly stated otherwise all files in this repository are licensed
# under the BSD-3-Clause License. This product includes software developed at
# Datadog (https://www.datadoghq.com/).

# Copyright 2021-Present Datadog, Inc

import ctypes
import os
import shutil
import socket
import sys
import tempfile
import threading
import time
from multiprocessing import Array, Event, Process, Value

# pylint: disable=too-many-instance-attributes,useless-object-inheritance
class FakeServer(object):
    """
    Fake statsd server that can be used for testing/benchmarking. Implementation
    Uses a separate process to run and manage the context to not poison the
    benchmarking results.
    """

    SOCKET_NAME = "fake_statsd_server_socket"
    ALLOWED_TRANSPORTS = ["UDS", "UDP"]
    MIN_RECV_BUFFER_SIZE = 32 * 1024

    def __init__(self, transport="UDS", ignore_timeouts=True, debug=False):
        if transport not in self.ALLOWED_TRANSPORTS:
            raise ValueError(
                "Transport {} is not a valid transport type. Only {} are allowed!".format(
                    transport,
                    self.ALLOWED_TRANSPORTS,
                )
            )

        self.transport = transport
        self.ignore_timeouts = ignore_timeouts
        self.debug = debug

        self.server_process = None
        self.socket_dir = None

        # Inter-process coordination events
        self.exit = Event()
        self.ready = Event()

        # Shared-mem property value holders for inter-process communication
        self._socket_path = Array(ctypes.c_char, 1024, lock=True)
        self._port = Value(ctypes.c_long, 0, lock=True)
        self._metric_counter_shmem_var = Value(ctypes.c_long, 0, lock=True)
        self._payload_counter_shmem_var = Value(ctypes.c_long, 0, lock=True)

    def _run_server(self):
        payload_counter = 0
        metric_counter = 0

        if self.transport == "UDS":
            self.socket_dir = tempfile.mkdtemp(prefix=self.__class__.__name__)
            socket_path = os.path.join(self.socket_dir, self.SOCKET_NAME)

            if os.path.exists(socket_path):
                os.unlink(socket_path)

            sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            sock.settimeout(3)

            # Increase the receiving buffer size where needed (e.g. MacOS has 4k RX
            # buffers which is half of the max packet size that the client will send.
            if os.name != 'nt':
                recv_buff_size = sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
                if recv_buff_size <= self.MIN_RECV_BUFFER_SIZE:
                    sock.setsockopt(
                        socket.SOL_SOCKET,
                        socket.SO_RCVBUF,
                        self.MIN_RECV_BUFFER_SIZE,
                    )

            sock.bind(socket_path)

            if self.debug:
                print("Listening via UDS on", socket_path)

            # We are using ctypes for shmem so we have to use a consistent
            # datatype across Python versions
            if sys.version_info[0] > 2:
                self._socket_path.value = socket_path.encode("utf-8")
            else:
                self._socket_path.value = socket_path

        elif self.transport == "UDP":
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(3)
            sock.bind(("", 0))

            _, self._port.value = sock.getsockname()

            if self.debug:
                print("Listening via UDP on port", self.port)

        # Run an async thread to update our shared-mem counters. We don't want to update
        # the shared memory values when we get data due to performance reasons.
        def _update_counters():
            while not self.exit.is_set():
                self._payload_counter_shmem_var.value = payload_counter
                self._metric_counter_shmem_var.value = metric_counter
                time.sleep(0.2)

        counter_update_timer = threading.Thread(target=_update_counters)
        counter_update_timer.daemon = True
        counter_update_timer.start()

        try:
            self.ready.set()

            while not self.exit.is_set():
                try:
                    payload, _ = sock.recvfrom(8192)
                except socket.timeout as ste:
                    if self.ignore_timeouts is True:
                        continue

                    raise ste

                payload_counter += 1

                offset = 0
                if payload[-1] == b"\n":
                    offset = -1

                metric_counter += len(payload[:offset].split(b"\n"))

                if self.debug:
                    print(
                        "Got '{}' (pkts: {}, payloads: {}, metrics: {})".format(
                            payload.decode('utf-8'),
                            len(payload[:offset].split(b"\n")),
                            payload_counter,
                            metric_counter,
                        )
                    )

        except socket.timeout as ste:
            if not self.exit.is_set():
                self.exit.set()
                raise ste
        finally:
            counter_update_timer.join()
            sock.close()

    def __enter__(self):
        if self.server_process:
            raise RuntimeError("Server already running")

        self.server_process = Process(
            target=self._run_server,
            name=FakeServer.__class__.__name__,
            args=(),
        )
        self.server_process.daemon = True
        self.server_process.start()

        self.ready.wait(5)

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        # Allow grace time to capture all metrics
        time.sleep(2)

        self.exit.set()
        self.server_process.join(10)

        if self.socket_dir:
            shutil.rmtree(self.socket_dir, ignore_errors=True)

        if self.server_process.exitcode != 0:
            raise RuntimeError("Server process did not exit successfully!")

    @property
    def port(self):
        return self._port.value

    @property
    def socket_path(self):
        return self._socket_path.value or None

    @property
    def payloads_captured(self):
        return self._payload_counter_shmem_var.value

    @property
    def metrics_captured(self):
        return self._metric_counter_shmem_var.value

    def __repr__(self):
        return "<FakeThreadedUDPServer(Packets RX: {}, Metrics RX: {}".format(
            self._payload_counter_shmem_var.value, self._metric_counter_shmem_var.value
        )


if __name__ == '__main__':
    options = {
      'ignore_timeouts': True,
      'debug': True,
    }

    if len(sys.argv) > 1:
        options['transport'] = sys.argv[1].upper()

    with FakeServer(**options) as server:
        while not server.exit.is_set():
            time.sleep(0.5)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
from collections import deque
from functools import reduce
import threading
import time
import unittest

# 3p
from mock import patch

# datadog
from datadog.dogstatsd.base import DogStatsd
from datadog.util.compat import is_p3k


class FakeSocket(object):
    """
    Mocked socket for testing.
    """
    def __init__(self):
        self.payloads = deque()

    def send(self, payload):
        if is_p3k():
            assert type(payload) == bytes
        else:
            assert type(payload) == str

        self.payloads.append(payload)

    def recv(self):
        try:
            return self.payloads
        except IndexError:
            return None

    def __repr__(self):
        return str(self.payloads)


class TestDogStatsdThreadSafety(unittest.TestCase):
    """
    DogStatsd thread safety tests.
    """
    def setUp(self):
        """
        Mock a socket.
        """
        self.socket = FakeSocket()

    def assertMetrics(self, values):
        """
        Helper, assertions on metrics.
        """
        count = len(values)

        # Split packet per metric (required when buffered) and discard empty packets
        packets = map(lambda x: x.split(b"\n"), self.socket.recv())
        packets = reduce(lambda prev, ele: prev + ele, packets, [])
        packets = list(filter(lambda x: x, packets))

        # Count
        self.assertEqual(
            len(packets), count,
            u"Metric size assertion failed: expected={expected}, received={received}".format(
                expected=count, received=len(packets)
            )
        )
        # Values
        for packet in packets:
            metric_value = int(packet.split(b':', 1)[1].split(b'|', 1)[0])
            self.assertIn(
                metric_value, values,
                u"Metric assertion failed: unexpected metric value {metric_value}".format(
                    metric_value=metric_value
                )
            )
            values.remove(metric_value)

    def test_socket_creation(self):
        """
        Socket creation plays well with multiple threads.
        """
        # Create a DogStatsd client but no socket
        statsd = DogStatsd()

        # Submit metrics from different threads to create a socket
        threads = []
        for value in range(10000):
            t = threading.Thread(target=statsd.gauge, args=("foo", value))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()

    @staticmethod
    def _submit_with_multiple_threads(statsd, submit_method, values):
        """
        Helper, use the given statsd client and method to submit the values
        within multiple threads.
        """
        threads = []
        for value in values:
            t = threading.Thread(
                target=getattr(statsd, submit_method),
                args=("foo", value)
            )
            threads.append(t)
            t.start()
        for t in threads:
            t.join()

    def test_increment(self):
        """
        Increments can be submitted from concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Samples
        values = set(range(10000))

        # Submit metrics from different threads
        self._submit_with_multiple_threads(statsd, "increment", values)

        #  All metrics were properly submitted
        self.assertMetrics(values)

    def test_decrement(self):
        """
        Decrements can be submitted from concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Samples
        values = set(range(10000))
        expected_value = set([-value for value in values])

        # Submit metrics from different threads
        self._submit_with_multiple_threads(statsd, "decrement", expected_value)

        #  All metrics were properly submitted
        self.assertMetrics(values)

    def test_gauge(self):
        """
        Gauges can be submitted from concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Samples
        values = set(range(10000))

        # Submit metrics from different threads
        self._submit_with_multiple_threads(statsd, "gauge", values)

        #  All metrics were properly submitted
        self.assertMetrics(values)

    def test_histogram(self):
        """
        Histograms can be submitted from concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Samples
        values = set(range(10000))

        # Submit metrics from different threads
        self._submit_with_multiple_threads(statsd, "histogram", values)

        #  All metrics were properly submitted
        self.assertMetrics(values)

    def test_timing(self):
        """
        Timings can be submitted from concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Samples
        values = set(range(10000))

        # Submit metrics from different threads
        self._submit_with_multiple_threads(statsd, "timing", values)

        # All metrics were properly submitted
        self.assertMetrics(values)

    def test_send_batch_metrics(self):
        """
        Metrics can be buffered, submitted from concurrent threads.
        """
        with DogStatsd() as batch_statsd:
            # Create a DogStatsd buffer client with a mocked socket
            batch_statsd.socket = self.socket

            # Samples
            values = set(range(10000))

            # Submit metrics from different threads
            self._submit_with_multiple_threads(batch_statsd, "gauge", values)

        # All metrics were properly submitted
        self.assertMetrics(values)

    @patch('datadog.dogstatsd.context.monotonic')
    def test_timed_decorator_threaded(self, mock_monotonic):
        """
        `timed` decorator plays well with concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Set up the mocked time
        mock_monotonic.return_value = 0

        # Method to time
        @statsd.timed("foo")
        def bar():
            """
            Wait 5 time units and return.
            """
            initial_time = mock_monotonic.return_value
            while mock_monotonic.return_value < initial_time + 2:
                pass

        # Run the method within multiple threads
        threads = []
        for value in range(10):
            t = threading.Thread(target=bar)
            threads.append(t)
            # Bump time so that previous thread can complete
            mock_monotonic.return_value += 1
            t.start()
            # Sleep to let the threads start
            time.sleep(0.1)

        # Bump time so that all threads completes
        time.sleep(0.1)
        mock_monotonic.return_value += 1
        time.sleep(0.1)
        mock_monotonic.return_value += 1

        for t in threads:
            t.join()

        # All metrics were properly submitted
        expected_values = [2 for _ in range(0, 10)]
        self.assertMetrics(expected_values)

    @patch('datadog.dogstatsd.context.monotonic')
    def test_timed_context_manager_threaded(self, mock_monotonic):
        """
        `timed` context manager plays well with concurrent threads.
        """
        # Create a DogStatsd client with a mocked socket
        statsd = DogStatsd()
        statsd.socket = self.socket

        # Set up the mocked time
        mock_monotonic.return_value = 0

        # Method to time
        def bar():
            """
            Wait 5 time units and return.
            """
            initial_time = mock_monotonic.return_value

            with statsd.timed("foo"):
                while mock_monotonic.return_value < initial_time + 2:
                    pass

        # Run the method within multiple threads
        threads = []
        for value in range(10):
            t = threading.Thread(target=bar)
            threads.append(t)
            # Bump time so that previous thread can complete
            mock_monotonic.return_value += 1
            t.start()
            # Sleep to let the threads start
            time.sleep(0.1)

        # Bump time so that all threads completes
        time.sleep(0.1)
        mock_monotonic.return_value += 1
        time.sleep(0.1)
        mock_monotonic.return_value += 1

        for t in threads:
            t.join()

        # All metrics were properly submitted
        expected_values = [2 for _ in range(0, 10)]
        self.assertMetrics(expected_values)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc.

from setuptools import setup

from io import open
import sys


def get_readme_md_contents():
    """read the contents of your README file"""
    with open("README.md", encoding='utf-8') as f:
        long_description = f.read()
        return long_description

version = {}
with open("datadog/version.py") as fp:
    exec(fp.read(), version)

# `typing` package is only required for Python versions older than 3.5, but we include it here so
# that a single wheel build can be used for all supported Python versions. Installing `typing` on
# Python 3.5+ has no effect.
#
# `configparser` package is only required for Python versions older than 3 (it is included here for
# the same reason as the `typing` package).
install_reqs = [
    "requests>=2.6.0",
    'typing;python_version<"3.5"',
    'configparser<5;python_version<"3.0"',
]

setup(
    name="datadog",
    version=version["__version__"],
    install_requires=install_reqs,
    tests_require=["pytest", "mock", "freezegun", "psutil"],
    packages=["datadog", "datadog.api", "datadog.dogstatsd", "datadog.threadstats", "datadog.util", "datadog.dogshell"],
    package_data={"datadog": ["py.typed"]},
    author="Datadog, Inc.",
    long_description=get_readme_md_contents(),
    long_description_content_type="text/markdown",
    author_email="dev@datadoghq.com",
    description="The Datadog Python library",
    license="BSD",
    keywords="datadog",
    url="https://www.datadoghq.com",
    project_urls={
        "Bug Tracker": "https://github.com/DataDog/datadogpy/issues",
        "Documentation": "https://datadogpy.readthedocs.io/en/latest/",
        "Source Code": "https://github.com/DataDog/datadogpy",
    },
    entry_points={
        "console_scripts": [
            "dog = datadog.dogshell:main",
            "dogwrap = datadog.dogshell.wrap:main",
            "dogshell = datadog.dogshell:main",
            "dogshellwrap = datadog.dogshell.wrap:main"
        ]
    },
    test_suite="tests",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # Required by Mypy when declaring PEP 561 compatibility with `py.typed`
    # file.
    zip_safe=False
)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import pytest


class TestAzureIntegration:

    test_tenant_name = "testc44-1234-5678-9101-cc00736ftest"
    test_client_id = "testc7f6-1234-5678-9101-3fcbf464test"
    test_client_secret = "testingx./Sw*g/Y33t..R1cH+hScMDt"
    test_new_tenant_name = "1234abcd-1234-5678-9101-abcd1234abcd"
    test_new_client_id = "abcd1234-5678-1234-5678-1234abcd5678"
    not_yet_installed_error = 'Azure Integration not yet installed.'

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def azure_integration(self, dog):
        """Prepare Azure Integration."""
        yield
        # Should be deleted as part of the test
        # but cleanup here if test fails
        dog.AzureIntegration.delete(
            tenant_name=self.test_new_tenant_name,
            client_id=self.test_new_client_id
        )

    def test_azure_crud(self, dog):
        # Test Create
        create_output = dog.AzureIntegration.create(
            tenant_name=self.test_tenant_name,
            host_filters="api:test",
            client_id=self.test_client_id,
            client_secret=self.test_client_secret
        )
        assert create_output == {}
        # Test List
        list_tests_pass = False
        for i in dog.AzureIntegration.list():
            if (i['tenant_name'] == self.test_tenant_name and
                    i['host_filters'] == 'api:test'):
                list_tests_pass = True
        assert list_tests_pass
        # Test Update Host Filters
        dog.AzureIntegration.update_host_filters(
            tenant_name=self.test_tenant_name,
            host_filters='api:test2',
            client_id=self.test_client_id
        )
        update_host_filters_tests_pass = False
        for i in dog.AzureIntegration.list():
            if i['host_filters'] == 'api:test2':
                update_host_filters_tests_pass = True
        assert update_host_filters_tests_pass
        # Test Update
        dog.AzureIntegration.update(
            tenant_name=self.test_tenant_name,
            new_tenant_name=self.test_new_tenant_name,
            host_filters="api:test3",
            client_id=self.test_client_id,
            new_client_id=self.test_new_client_id,
            client_secret=self.test_client_secret
        )
        update_tests_pass = False
        for i in dog.AzureIntegration.list():
            if (i['tenant_name'] == self.test_new_tenant_name and
                    i['host_filters'] == 'api:test3'):
                update_tests_pass = True
        assert update_tests_pass
        # Test Delete
        dog.AzureIntegration.delete(
            tenant_name=self.test_new_tenant_name,
            client_id=self.test_new_client_id
        )
        delete_tests_pass = True
        list_output = dog.AzureIntegration.list()
        if type(list_output) == list:
            for i in dog.AzureIntegration.list():
                if i['tenant_name'] == self.test_new_tenant_name:
                    delete_tests_pass = False
        elif self.not_yet_installed_error in list_output['errors'][0]:
            pass
        assert delete_tests_pass



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import os
from contextlib import contextmanager


@contextmanager
def preserve_environment_variable(env_name):
    environ_api_param = os.environ.get(env_name)
    try:
        yield
    finally:
        if environ_api_param is not None:
            os.environ[env_name] = environ_api_param
        else:
            del os.environ[env_name]


# Code copied from - https://github.com/DataDog/integrations-core/blob/de1b684e4e98d06a7b0da3249805de74bb877cea/datadog_checks_dev/datadog_checks/dev/structures.py#L24
class EnvVars(dict):
    def __init__(self, env_vars=None, ignore=None):
        super(EnvVars, self).__init__(os.environ)
        self.old_env = dict(self)

        if env_vars is not None:
            self.update(env_vars)

        if ignore is not None:
            for env_var in ignore:
                self.pop(env_var, None)

    def __enter__(self):
        os.environ.clear()
        os.environ.update(self)

    def __exit__(self, exc_type, exc_value, traceback):
        os.environ.clear()
        os.environ.update(self.old_env)



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""Record HTTP requests to avoid hiting Datadog API from CI."""

import logging
import os
import time
from datetime import datetime

import pytest
from vcr import VCR
from dateutil import parser

from tests.integration.api.constants import API_KEY, APP_KEY, API_HOST, TEST_USER

WAIT_TIME = 10
FAKE_PROXY = {"https": "http://user:pass@10.10.1.10:3128/"}

logging.basicConfig()
vcr_log = logging.getLogger("vcr")
vcr_log.setLevel(logging.INFO)


@pytest.fixture(scope="module")
def api():
    """Initialize Datadog API client."""
    from datadog import api, initialize
    from datadog.api.api_client import APIClient
    APIClient._sort_keys = True
    initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST)

    http_client = APIClient._get_http_client()
    try:
        assert http_client._session is None
        http_client.request(None, None, None, None, None, None, None, None, max_retries=10)
    except Exception as e:
        assert http_client._session is not None

    return api


@pytest.fixture(scope='module')
def vcr_config():
    return dict(
        filter_headers=('DD-API-KEY', 'DD-APPLICATION-KEY'),
        filter_query_parameters=('api_key', 'application_key'),
    )


@pytest.fixture
def freezer(vcr_cassette_name, vcr_cassette, vcr):
    from freezegun import freeze_time

    if vcr_cassette.record_mode == "all":
        tzinfo = datetime.now().astimezone().tzinfo
        freeze_at = datetime.now().replace(tzinfo=tzinfo).isoformat()
        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".frozen"
            ),
            "w",
        ) as f:
            f.write(freeze_at)
    else:
        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".frozen"
            ),
            "r",
        ) as f:
            freeze_at = f.readline().strip()

    freeze_at = parser.isoparse(freeze_at)

    # dt = parser.isoparse(freeze_at)
    # tz_offset = dt.tzinfo.utcoffset(dt).seconds / 60
    # os.environ['TZ'] =  "UTC%+03d:%02d" % (
    #     int( tz_offset / 60), tz_offset % 60
    # )
    # time.tzset()

    return freeze_time(freeze_at)


@pytest.fixture
def user_handle(vcr_cassette_name, vcr_cassette, vcr):
    if vcr_cassette.record_mode == "all":
        assert TEST_USER is not None, "You must set DD_TEST_CLIENT_USER environment variable to run comment tests"
        handle = TEST_USER
        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".handle"
            ),
            "w",
        ) as f:
            f.write(handle)
    else:
        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".handle"
            ),
            "r",
        ) as f:
            handle = f.readline().strip()

    return handle


@pytest.fixture
def dog(api, vcr_cassette):
    """Record communication with Datadog API."""
    from datadog.util.compat import is_p3k
    if not is_p3k() and vcr_cassette.record_mode != "all":
        pytest.skip("Can not replay responses on Python 2")

    old_host_name = api._host_name
    api._host_name = "test.host"

    yield api

    api._host_name = old_host_name


@pytest.fixture
def get_with_retry(vcr_cassette, dog):
    """Return a retry factory that correctly handles the request recording."""

    def retry(
            resource_type,
            resource_id=None,
            operation="get",
            retry_limit=10,
            retry_condition=lambda r: r.get("errors"),
            **kwargs
    ):
        number_of_interactions = len(vcr_cassette.data) if vcr_cassette.record_mode == "all" else -1

        if resource_id is None:
            resource = getattr(getattr(dog, resource_type), operation)(**kwargs)
        else:
            resource = getattr(getattr(dog, resource_type), operation)(resource_id, **kwargs)
        retry_counter = 0
        while retry_condition(resource) and retry_counter < retry_limit:
            time.sleep(WAIT_TIME)

            if vcr_cassette.record_mode == "all":
                # remove failed interactions
                vcr_cassette.data = vcr_cassette.data[:number_of_interactions]

            if resource_id is None:
                resource = getattr(getattr(dog, resource_type), operation)(**kwargs)
            else:
                resource = getattr(getattr(dog, resource_type), operation)(resource_id, **kwargs)
            retry_counter += 1

        if retry_condition(resource):
            raise Exception(
                "Retry limit reached performing `{}` on resource {}, ID {}".format(operation, resource_type, resource_id)
            )
        return resource
    return retry



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import pytest


class TestGcpIntegration:

    test_project_id = "datadog-apitest"
    test_client_email = "api-dev@datadog-sandbox.iam.gserviceaccount.com"

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def gcp_integration(self, dog):
        """Prepare GCP Integration."""
        yield
        # Should be deleted as part of the test
        # but cleanup here if test fails
        dog.GcpIntegration.delete(
            project_id=self.test_project_id,
            client_email=self.test_client_email
        )

    def test_gcp_crud(self, dog):
        # Test Create
        create_output = dog.GcpIntegration.create(
            type="service_account",
            project_id=self.test_project_id,
            private_key_id="fake_private_key_id",
            private_key="fake_key",
            client_email=self.test_client_email,
            client_id="123456712345671234567",
            auth_uri="fake_uri",
            token_uri="fake_uri",
            auth_provider_x509_cert_url="fake_url",
            client_x509_cert_url="fake_url",
            host_filters="api:test"
        )
        assert create_output == {}
        # Test Update
        dog.GcpIntegration.update(
            project_id=self.test_project_id,
            client_email=self.test_client_email,
            host_filters="api:test2",
            automute=True
        )
        update_tests_pass = False
        for i in dog.GcpIntegration.list():
            if (i['project_id'] == self.test_project_id and
                    i['host_filters'] == 'api:test2' and
                    i['automute'] is True):
                update_tests_pass = True
        assert update_tests_pass
        # Test List
        list_tests_pass = False
        for i in dog.GcpIntegration.list():
            if (i['project_id'] == self.test_project_id and
                    i['host_filters'] == 'api:test2' and
                    i['automute'] is True):
                list_tests_pass = True
        assert list_tests_pass
        # Test Delete
        delete_output = dog.GcpIntegration.delete(
            project_id=self.test_project_id,
            client_email=self.test_client_email
        )
        assert delete_output == {}



# coding: utf8
# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import unittest

import pytest

from datadog.util.format import construct_url, normalize_tags


class TestConstructURL:
    expected = "https://api.datadoghq.com/api/v1/graph/snapshot"
    test_data = [
        ("https://api.datadoghq.com", "v1", "graph/snapshot", expected),
        ("https://api.datadoghq.com/", "v1", "graph/snapshot", expected),
        ("https://api.datadoghq.com", "/v1", "graph/snapshot", expected),
        ("https://api.datadoghq.com/", "/v1", "graph/snapshot", expected),
        ("https://api.datadoghq.com", "v1/", "graph/snapshot", expected),
        ("https://api.datadoghq.com/", "v1/", "graph/snapshot", expected),
        ("https://api.datadoghq.com", "/v1/", "graph/snapshot", expected),
        ("https://api.datadoghq.com/", "/v1/", "graph/snapshot", expected),
        ("https://api.datadoghq.com", "v1", "/graph/snapshot", expected),
        ("https://api.datadoghq.com/", "v1", "/graph/snapshot", expected),
        ("https://api.datadoghq.com", "/v1", "/graph/snapshot", expected),
        ("https://api.datadoghq.com/", "/v1", "/graph/snapshot", expected),
        ("https://api.datadoghq.com", "v1/", "/graph/snapshot", expected),
        ("https://api.datadoghq.com/", "v1/", "/graph/snapshot", expected),
        ("https://api.datadoghq.com", "/v1/", "/graph/snapshot", expected),
        ("https://api.datadoghq.com/", "/v1/", "/graph/snapshot", expected),
    ]

    @pytest.mark.parametrize("host,api_version,path,expected", test_data)
    def test_construct_url(self, host, api_version, path, expected):
        assert construct_url(host, api_version, path) == expected

class TestNormalizeTags:
    """
    Test of the format's `normalize_tags` functionality
    """
    test_data = [
        ([], []),
        ([''],['']),
        (['this is a tag'], ['this_is_a_tag']),
        (['abc!@#$%^&*()0987654321{}}{'], ['abc__________0987654321____']),
        (['abc!@#', '^%$#3456#'], ['abc___', '____3456_']),
        (['mutliple', 'tags', 'included'], ['mutliple', 'tags', 'included']),
        ([u'абвгдежзийкл', u'абв' , 'test123'], [u'абвгдежзийкл', u'абв' , 'test123']),
        ([u'абвгд西😃ежзийкл', u'аб😃西в' , u'a😃😃b'],  [u'абвгд西_ежзийкл', u'аб_西в', u'a__b']),
    ]

    @pytest.mark.parametrize("original_tags,expected_tags", test_data)
    def test_normalize_tags(self, original_tags, expected_tags):
            assert normalize_tags(original_tags) == expected_tags



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
from copy import deepcopy
import json
import os
import tempfile
from time import time
import zlib

# 3p
import mock, pytest

# datadog
from datadog import initialize, api, util
from datadog.api import (
    Distribution,
    Event,
    Logs,
    Metric,
    ServiceCheck,
    User
)
from datadog.api.exceptions import (
    DatadogException,
    ProxyError,
    ClientError,
    HttpTimeout,
    HttpBackoff,
    HTTPError,
    ApiError,
    ApiNotInitialized,
)
from datadog.util.compat import is_p3k
from datadog.util.format import normalize_tags
from tests.unit.api.helper import (
    DatadogAPIWithInitialization,
    DatadogAPINoInitialization,
    MyCreatable,
    MyParamsApiKeyCreatable,
    MyUpdatable,
    MyDeletable,
    MyGetable,
    MyListable,
    MyListableSubResource,
    MyAddableSubResource,
    MyUpdatableSubResource,
    MyDeletableSubResource,
    MyActionable,
    API_KEY,
    APP_KEY,
    API_HOST,
    HOST_NAME,
    FAKE_PROXY
)
from datadog.util.hostname import CfgNotFound, get_hostname

from tests.util.contextmanagers import EnvVars


class TestInitialization(DatadogAPINoInitialization):

    def test_default_settings_set(self):
        """
        Test all the default setting are properly set before calling initialize
        """
        from datadog.api import (
            _api_key,
            _application_key,
            _api_version,
            _api_host,
            _host_name,
            _hostname_from_config,
            _cacert,
            _proxies,
            _timeout,
            _max_timeouts,
            _max_retries,
            _backoff_period,
            _mute,
            _return_raw_response,
        )

        assert _api_key is None
        assert _application_key is None
        assert _api_version == 'v1'
        assert _api_host is None
        assert _host_name is None
        assert _hostname_from_config is True
        assert _cacert is True
        assert _proxies is None
        assert _timeout == 60
        assert _max_timeouts == 3
        assert _max_retries == 3
        assert _backoff_period == 300
        assert _mute is True
        assert _return_raw_response is False

    def test_no_initialization_fails(self):
        """
        Raise ApiNotInitialized exception when `initialize` has not ran or no API key was set.
        """
        self.assertRaises(ApiNotInitialized, MyCreatable.create)

        # No API key => only stats in statsd mode should work
        initialize()
        api._api_key = None
        self.assertRaises(ApiNotInitialized, MyCreatable.create)

        # Finally, initialize with an API key
        initialize(api_key=API_KEY, api_host=API_HOST)
        MyCreatable.create()
        self.assertEqual(self.request_mock.call_count(), 1)

    @mock.patch('datadog.util.config.get_config_path')
    def test_get_hostname(self, mock_config_path):
        """
        API hostname parameter fallback with Datadog Agent hostname when available.
        """
        # Generate a fake agent config
        tmpfilepath = os.path.join(tempfile.gettempdir(), "tmp-agentconfig")
        with open(tmpfilepath, "wb") as f:
            if is_p3k():
                f.write(bytes("[Main]\n", 'UTF-8'))
                f.write(bytes("hostname: {0}\n".format(HOST_NAME), 'UTF-8'))
            else:
                f.write("[Main]\n")
                f.write("hostname: {0}\n".format(HOST_NAME))
        # Mock get_config_path to return this fake agent config
        mock_config_path.return_value = tmpfilepath

        initialize()
        self.assertEqual(api._host_name, HOST_NAME, api._host_name)

    def test_hostname_warning_not_present(self):
        try:
            get_hostname(hostname_from_config=False)
        except CfgNotFound:
            pytest.fail("Unexpected CfgNotFound Exception")

    def test_normalize_tags(self):
        tag_list_test = ["tag1, tag2", "tag3 ,tag4", "tag5,tag6"]
        tag_list_final = normalize_tags(tag_list_test)
        assert tag_list_final == ["tag1__tag2", "tag3__tag4", "tag5_tag6"]

    def test_errors_suppressed(self):
        """
        API `errors` field ApiError suppressed when specified
        """
        # Test API, application keys, API host, and some HTTP client options
        initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST)

        # Make a simple API call
        self.load_request_response(response_body='{"data": {}, "errors": ["foo error"]}')
        resp = MyCreatable.create(params={"suppress_response_errors_on_codes": [200]})
        self.assertNotIsInstance(resp, ApiError)
        self.assertDictEqual({"data": {}, "errors": ["foo error"]}, resp)

    def test_request_parameters(self):
        """
        API parameters are set with `initialize` method.
        """
        # Test API, application keys, API host, and some HTTP client options
        initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST)

        # Make a simple API call
        MyCreatable.create()

        _, options = self.request_mock.call_args()

        # Assert `requests` parameters
        self.assertIn('params', options)

        self.assertIn('headers', options)
        self.assertEqual(options['headers']['Content-Type'], 'application/json')
        self.assertEqual(options['headers']['DD-API-KEY'], API_KEY)
        self.assertEqual(options['headers']['DD-APPLICATION-KEY'], APP_KEY)
        assert "api_key" not in options['params']
        assert "application_key" not in options['params']

    def test_request_parameters_api_keys_in_params(self):
        """
        API parameters are set with `initialize` method.
        """
        # Test API, application keys, API host, and some HTTP client options
        initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST)

        # Make a simple API call
        MyParamsApiKeyCreatable.create()

        _, options = self.request_mock.call_args()

        # Assert `requests` parameters
        self.assertIn('params', options)

        self.assertIn('headers', options)

        # for resources in MyParamsApiKey, api key and application key needs to be in url params
        # any api and app keys in headers are ignored
        self.assertEqual(options['headers']['Content-Type'], 'application/json')
        self.assertEqual(options['params']['api_key'], API_KEY)
        self.assertEqual(options['params']['application_key'], APP_KEY)
        assert "DD-API-KEY" not in options['headers']
        assert "DD-APPLICATION-KEY" not in options['headers']

    def test_initialize_options(self):
        """
        HTTP client and API options are set with `initialize` method.
        """
        initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST,
                   proxies=FAKE_PROXY, cacert=False)

        # Make a simple API call
        MyCreatable.create()

        _, options = self.request_mock.call_args()

        # Assert `requests` parameters
        self.assertIn('proxies', options)
        self.assertEqual(options['proxies'], FAKE_PROXY)

        self.assertIn('verify', options)
        self.assertEqual(options['verify'], False)

        # Arm the `requests` to raise
        self.arm_requests_to_raise()

        # No exception should be raised (mute=True by default)
        MyCreatable.create()

        # Repeat with mute to False
        initialize(api_key=API_KEY, mute=False)
        self.assertRaises(ApiError, MyCreatable.create)

    def test_return_raw_response(self):
        # Test default initialization sets return_raw_response to False
        initialize()
        assert not api._return_raw_response
        # Assert that we can set this to True
        initialize(return_raw_response=True)
        assert api._return_raw_response
        # Assert we get multiple fields back when set to True
        initialize(api_key="aaaaaaaaaa", app_key="123456", return_raw_response=True)
        data, raw = api.Monitor.get_all()

    def test_default_values(self):
        with EnvVars(ignore=[
            "DATADOG_API_KEY",
            "DATADOG_APP_KEY",
            "DD_API_KEY",
            "DD_APP_KEY"
        ]):
            initialize()

            self.assertIsNone(api._api_key)
            self.assertIsNone(api._application_key)
            self.assertEqual(api._api_host, "https://api.datadoghq.com")
            self.assertEqual(api._host_name, util.hostname.get_hostname(api._hostname_from_config))

    def test_env_var_values(self):
        with EnvVars(
            env_vars={
                "DATADOG_API_KEY": "API_KEY_ENV",
                "DATADOG_APP_KEY": "APP_KEY_ENV",
                "DATADOG_HOST": "HOST_ENV",
            }
        ):
            initialize()

            self.assertEqual(api._api_key, "API_KEY_ENV")
            self.assertEqual(api._application_key, "APP_KEY_ENV")
            self.assertEqual(api._api_host, "HOST_ENV")
            self.assertEqual(api._host_name, util.hostname.get_hostname(api._hostname_from_config))

            del os.environ["DATADOG_API_KEY"]
            del os.environ["DATADOG_APP_KEY"]
            del os.environ["DATADOG_HOST"]

            with EnvVars(env_vars={
                "DD_API_KEY": "API_KEY_ENV_DD",
                "DD_APP_KEY": "APP_KEY_ENV_DD",
            }):
                api._api_key = None
                api._application_key = None

                initialize()

                self.assertEqual(api._api_key, "API_KEY_ENV_DD")
                self.assertEqual(api._application_key, "APP_KEY_ENV_DD")

    def test_function_param_value(self):
        initialize(api_key="API_KEY", app_key="APP_KEY", api_host="HOST", host_name="HOSTNAME")

        self.assertEqual(api._api_key, "API_KEY")
        self.assertEqual(api._application_key, "APP_KEY")
        self.assertEqual(api._api_host, "HOST")
        self.assertEqual(api._host_name, "HOSTNAME")

    def test_precedence(self):
        # Initialize first with env vars
        with EnvVars(env_vars={
            "DD_API_KEY": "API_KEY_ENV_DD",
            "DD_APP_KEY": "APP_KEY_ENV_DD",
        }):
            os.environ["DATADOG_API_KEY"] = "API_KEY_ENV"
            os.environ["DATADOG_APP_KEY"] = "APP_KEY_ENV"
            os.environ["DATADOG_HOST"] = "HOST_ENV"

            initialize()

            self.assertEqual(api._api_key, "API_KEY_ENV")
            self.assertEqual(api._application_key, "APP_KEY_ENV")
            self.assertEqual(api._api_host, "HOST_ENV")
            self.assertEqual(api._host_name, util.hostname.get_hostname(api._hostname_from_config))

            # Initialize again to check given parameters take precedence over already set value and env vars
            initialize(api_key="API_KEY", app_key="APP_KEY", api_host="HOST", host_name="HOSTNAME")

            self.assertEqual(api._api_key, "API_KEY")
            self.assertEqual(api._application_key, "APP_KEY")
            self.assertEqual(api._api_host, "HOST")
            self.assertEqual(api._host_name, "HOSTNAME")

            # Initialize again without specifying attributes to check that already initialized value takes precedence
            initialize()

            self.assertEqual(api._api_key, "API_KEY")
            self.assertEqual(api._application_key, "APP_KEY")
            self.assertEqual(api._api_host, "HOST")
            self.assertEqual(api._host_name, "HOSTNAME")

            del os.environ["DATADOG_API_KEY"]
            del os.environ["DATADOG_APP_KEY"]
            del os.environ["DATADOG_HOST"]


class TestExceptions(DatadogAPINoInitialization):

    def test_base_exception(self):
        args = [ "foo" ]
        with pytest.raises(DatadogException):
            raise DatadogException(*args)

    def test_proxyerror_exception(self):
        args = [ "GET", "http://localhost:8080", HTTPError("oh no") ]
        kwargs = { "method": "GET", "url": "http://localhost:8080", "exception": HTTPError("oh no") }
        with pytest.raises(ProxyError):
            raise ProxyError(*args)
        with pytest.raises(DatadogException):
            raise ProxyError(*args)
        with pytest.raises(ProxyError):
            raise ProxyError(**kwargs)
        with pytest.raises(DatadogException):
            raise ProxyError(**kwargs)

    def test_clienterror_exception(self):
        args = [ "GET", "http://localhost:8080", HTTPError("oh no") ]
        kwargs = { "method": "GET", "url": "http://localhost:8080", "exception": HTTPError("oh no") }
        with pytest.raises(ClientError):
            raise ClientError(*args)
        with pytest.raises(DatadogException):
            raise ClientError(*args)
        with pytest.raises(ClientError):
            raise ClientError(**kwargs)
        with pytest.raises(DatadogException):
            raise ClientError(**kwargs)

    def test_httptimeout_exception(self):
        args = [ "GET", "http://localhost:8080", 5 ]
        kwargs = { "method": "GET", "url": "http://localhost:8080", "timeout": 5 }
        with pytest.raises(HttpTimeout):
            raise HttpTimeout(*args)
        with pytest.raises(DatadogException):
            raise HttpTimeout(*args)
        with pytest.raises(HttpTimeout):
            raise HttpTimeout(**kwargs)
        with pytest.raises(DatadogException):
            raise HttpTimeout(**kwargs)

    def test_httpbackoff_exception(self):
        args = [ 30 ]
        kwargs = { "backoff_period": 30 }
        with pytest.raises(HttpBackoff):
            raise HttpBackoff(*args)
        with pytest.raises(DatadogException):
            raise HttpBackoff(*args)
        with pytest.raises(HttpBackoff):
            raise HttpBackoff(**kwargs)
        with pytest.raises(DatadogException):
            raise HttpBackoff(**kwargs)

    def test_httperror_exception(self):
        args = [ 500, "oh no" ]
        kwargs = { "status_code": 500, "reason": "oh no" }
        with pytest.raises(HTTPError):
            raise HTTPError(*args)
        with pytest.raises(DatadogException):
            raise HTTPError(*args)
        with pytest.raises(HTTPError):
            raise HTTPError(**kwargs)
        with pytest.raises(DatadogException):
            raise HTTPError(**kwargs)

    def test_apierror_exception(self):
        with pytest.raises(ApiError):
            raise ApiError()
        with pytest.raises(DatadogException):
            raise ApiError()

    def test_apinotinitialized_exception(self):
        with pytest.raises(ApiNotInitialized):
            raise ApiNotInitialized()
        with pytest.raises(DatadogException):
            raise ApiNotInitialized()


class TestResources(DatadogAPIWithInitialization):

    def test_creatable(self):
        """
        Creatable resource logic.
        """
        MyCreatable.create(mydata="val")
        self.request_called_with('POST', API_HOST + "/api/v1/creatables", data={'mydata': "val"})

        MyCreatable.create(mydata="val", attach_host_name=True)
        self.request_called_with('POST', API_HOST + "/api/v1/creatables",
                                 data={'mydata': "val", 'host': api._host_name})

    def test_getable(self):
        """
        Getable resource logic.
        """
        getable_object_id = 123
        MyGetable.get(getable_object_id, otherparam="val")
        self.request_called_with('GET', API_HOST + "/api/v1/getables/" + str(getable_object_id),
                                 params={'otherparam': "val"})
        _, kwargs = self.request_mock.call_args()
        self.assertIsNone(kwargs["data"])

    def test_listable(self):
        """
        Listable resource logic.
        """
        MyListable.get_all(otherparam="val")
        self.request_called_with('GET', API_HOST + "/api/v1/listables", params={'otherparam': "val"})
        _, kwargs = self.request_mock.call_args()
        self.assertIsNone(kwargs["data"])

    def test_updatable(self):
        """
        Updatable resource logic.
        """
        updatable_object_id = 123
        MyUpdatable.update(updatable_object_id, params={'myparam': "val1"}, mydata="val2")
        self.request_called_with('PUT', API_HOST + "/api/v1/updatables/" + str(updatable_object_id),
                                 params={'myparam': "val1"}, data={'mydata': "val2"})

    def test_detalable(self):
        """
        Deletable resource logic.
        """
        deletable_object_id = 123
        MyDeletable.delete(deletable_object_id, otherparam="val")
        self.request_called_with('DELETE', API_HOST + "/api/v1/deletables/" + str(deletable_object_id),
                                 params={'otherparam': "val"})

    def test_listable_sub_resources(self):
        """
        Listable sub-resources logic.
        """
        resource_id = 123
        MyListableSubResource.get_items(resource_id, otherparam="val")
        self.request_called_with(
            'GET',
            API_HOST + '/api/v1/resource_name/{0}/sub_resource_name'.format(resource_id),
            params={'otherparam': "val"}
        )
        _, kwargs = self.request_mock.call_args()
        self.assertIsNone(kwargs["data"])

    def test_addable_sub_resources(self):
        """
        Addable sub-resources logic.
        """
        resource_id = 123
        MyAddableSubResource.add_items(resource_id, params={'myparam': 'val1'}, mydata='val2')
        self.request_called_with(
            'POST',
            API_HOST + '/api/v1/resource_name/{0}/sub_resource_name'.format(resource_id),
            params={'myparam': 'val1'},
            data={'mydata': 'val2'}
        )

    def test_updatable_sub_resources(self):
        """
        Updatable sub-resources logic.
        """
        resource_id = 123
        MyUpdatableSubResource.update_items(resource_id, params={'myparam': 'val1'}, mydata='val2')
        self.request_called_with(
            'PUT',
            API_HOST + '/api/v1/resource_name/{0}/sub_resource_name'.format(resource_id),
            params={'myparam': 'val1'},
            data={'mydata': 'val2'}
        )

    def test_deletable_sub_resources(self):
        """
        Deletable sub-resources logic.
        """
        resource_id = 123
        MyDeletableSubResource.delete_items(resource_id, params={'myparam': 'val1'}, mydata='val2')
        self.request_called_with(
            'DELETE',
            API_HOST + '/api/v1/resource_name/{0}/sub_resource_name'.format(resource_id),
            params={'myparam': 'val1'},
            data={'mydata': 'val2'}
        )

    def test_actionable(self):
        """
        Actionable resource logic.
        """
        actionable_object_id = 123
        MyActionable.trigger_class_action(
            'POST',
            'actionname',
            id=actionable_object_id,
            params={'myparam': 'val1'},
            mydata='val',
            mydata2='val2'
        )
        self.request_called_with(
            'POST',
            API_HOST + '/api/v1/actionables/{0}/actionname'.format(str(actionable_object_id)),
            params={'myparam': 'val1'},
            data={'mydata': 'val', 'mydata2': 'val2'}
        )

        MyActionable.trigger_class_action(
            'POST',
            'actionname',
            id=actionable_object_id,
            mydata='val',
            mydata2='val2'
        )
        self.request_called_with(
            'POST',
            API_HOST + '/api/v1/actionables/{0}/actionname'.format(str(actionable_object_id)),
            params={},
            data={'mydata': 'val', 'mydata2': 'val2'}
        )

        MyActionable.trigger_class_action(
            'GET',
            'actionname',
            id=actionable_object_id,
            params={'param1': 'val1', 'param2': 'val2'}
        )
        self.request_called_with(
            'GET',
            API_HOST + '/api/v1/actionables/{0}/actionname'.format(str(actionable_object_id)),
            params={'param1': 'val1', 'param2': 'val2'}
        )
        _, kwargs = self.request_mock.call_args()
        self.assertIsNone(kwargs["data"])

        MyActionable.trigger_action(
            'POST',
            'actionname',
            id=actionable_object_id,
            mydata="val"
        )
        self.request_called_with(
            'POST',
            API_HOST + '/api/v1/actionname/{0}'.format(actionable_object_id),
            data={'mydata': "val"}
        )

        MyActionable.trigger_action(
            'GET',
            'actionname',
            id=actionable_object_id,
        )
        self.request_called_with(
            'GET',
            API_HOST + '/api/v1/actionname/{0}'.format(actionable_object_id)
        )
        _, kwargs = self.request_mock.call_args()
        self.assertIsNone(kwargs["data"])


class TestEventResource(DatadogAPIWithInitialization):

    def test_submit_event_wrong_alert_type(self):
        """
        Assess that an event submitted with a wrong alert_type raises the correct Exception
        """
        with pytest.raises(ApiError) as excinfo:
            Event.create(
                title="test no hostname", text="test no hostname", attach_host_name=False, alert_type="wrong_type"
            )
        assert "Parameter alert_type must be either error, warning, info or success" in str(excinfo.value)


class TestLogsResource(DatadogAPIWithInitialization):
    def test_list_logs(self):
        Logs.list(data={"time": {"from": "2021-01-01T11:00:00Z", "to": "2021-01-02T11:00:00Z"}})
        self.request_called_with(
            "POST",
            "https://example.com/api/v1/logs-queries/list",
            data={"time": {"from": "2021-01-01T11:00:00Z", "to": "2021-01-02T11:00:00Z"}}
        )


class TestMetricResource(DatadogAPIWithInitialization):

    def submit_and_assess_metric_payload(self, serie, attach_host_name=True):
        """
        Helper to assess the metric payload format.
        """
        now = time()

        if isinstance(serie, dict):
            Metric.send(attach_host_name=attach_host_name, **deepcopy(serie))
            serie = [serie]
        else:
            Metric.send(deepcopy(serie), attach_host_name=attach_host_name)

        payload = self.get_request_data()

        for i, metric in enumerate(payload['series']):
            if attach_host_name:
                self.assertEqual(set(metric.keys()), set(['metric', 'points', 'host']))
                self.assertEqual(metric['host'], api._host_name)
            else:
                self.assertEqual(set(metric.keys()), set(['metric', 'points']))

            self.assertEqual(metric['metric'], serie[i]['metric'])

            # points is a list of 1 point
            self.assertTrue(isinstance(metric['points'], list))
            self.assertEqual(len(metric['points']), 1)
            # it consists of a [time, value] pair
            self.assertEqual(len(metric['points'][0]), 2)
            # its value == value we sent
            self.assertEqual(metric['points'][0][1], float(serie[i]['points']))
            # it's time not so far from current time
            assert now - 1 < metric['points'][0][0] < now + 1

    def submit_and_assess_dist_payload(self, serie, attach_host_name=True):
        """
        Helper to assess the metric payload format.
        """
        now = time()

        if isinstance(serie, dict):
            Distribution.send(attach_host_name=attach_host_name, **deepcopy(serie))
            serie = [serie]
        else:
            Distribution.send(deepcopy(serie), attach_host_name=attach_host_name)

        payload = self.get_request_data()

        for i, metric in enumerate(payload['series']):
            if attach_host_name:
                self.assertEqual(set(metric.keys()), set(['metric', 'points', 'host']))
                self.assertEqual(metric['host'], api._host_name)
            else:
                self.assertEqual(set(metric.keys()), set(['metric', 'points']))

            self.assertEqual(metric['metric'], serie[i]['metric'])

            # points is a list of 1 point
            self.assertTrue(isinstance(metric['points'], list))
            self.assertEqual(len(metric['points']), 1)
            # it consists of a [time, value] pair
            self.assertEqual(len(metric['points'][0]), 2)
            # its value == value we sent
            self.assertEqual(metric['points'][0][1], serie[i]['points'][0][1])
            # it's time not so far from current time
            assert now - 1 < metric['points'][0][0] < now + 1

    def test_metric_submit_query_switch(self):
        """
        Endpoints are different for submission and queries.
        """
        Metric.send(points=(123, 456))
        self.request_called_with('POST', API_HOST + "/api/v1/series",
                                 data={'series': [{'points': [[123, 456.0]], 'host': api._host_name}]})

        Metric.query(start="val1", end="val2")
        self.request_called_with('GET', API_HOST + "/api/v1/query",
                                 params={'from': "val1", 'to': "val2"})

    def test_points_submission(self):
        """
        Assess the data payload format, when submitting a single or multiple points.
        """
        # Single point
        serie = dict(metric='metric.1', points=13)
        self.submit_and_assess_metric_payload(serie)

        # Multiple point
        serie = [dict(metric='metric.1', points=13),
                 dict(metric='metric.2', points=19)]
        self.submit_and_assess_metric_payload(serie)

        # Single point no hostname
        serie = dict(metric='metric.1', points=13)
        self.submit_and_assess_metric_payload(serie, attach_host_name=False)

        # Multiple point no hostname
        serie = [dict(metric='metric.1', points=13),
                 dict(metric='metric.2', points=19)]
        self.submit_and_assess_metric_payload(serie, attach_host_name=False)

    def test_dist_points_submission(self):
        """
        Assess the distribution data payload format, when submitting a single or multiple points.
        """
        # Single point
        serie = dict(metric='metric.1', points=[[time(), [13]]])
        self.submit_and_assess_dist_payload(serie)

        # Multiple point
        serie = [dict(metric='metric.1', points=[[time(), [13]]]),
                 dict(metric='metric.2', points=[[time(), [19]]])]
        self.submit_and_assess_dist_payload(serie)

        # Single point no hostname
        serie = dict(metric='metric.1', points=[[time(), [13]]])
        self.submit_and_assess_dist_payload(serie, attach_host_name=False)

        # Multiple point no hostname
        serie = [dict(metric='metric.1', points=[[time(), [13]]]),
                 dict(metric='metric.2', points=[[time(), [19]]])]
        self.submit_and_assess_dist_payload(serie, attach_host_name=False)

    def test_data_type_support(self):
        """
        `Metric` API supports `real` numerical data types.
        """
        from decimal import Decimal
        from fractions import Fraction

        m_long = int(1)  # long in Python 3.x

        if not is_p3k():
            m_long = long(1)  # noqa: F821

        supported_data_types = [1, 1.0, m_long, Decimal(1), Fraction(1, 2)]

        for point in supported_data_types:
            serie = dict(metric='metric.numerical', points=point)
            self.submit_and_assess_metric_payload(serie)

    def test_compression(self):
        """
        Metric and Distribution support zlib compression
        """

        # By default, there is no compression
        # Metrics
        series = dict(metric="metric.1", points=[(time(), 13.)])
        Metric.send(attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" not in headers
        assert req_data == json.dumps({"series": [series]})
        # Same result when explicitely False
        Metric.send(compress_payload=False, attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" not in headers
        assert req_data == json.dumps({"series": [series]})
        # Distributions
        series = dict(metric="metric.1", points=[(time(), 13.)])
        Distribution.send(attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" not in headers
        assert req_data == json.dumps({"series": [series]})
        # Same result when explicitely False
        Distribution.send(compress_payload=False, attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" not in headers
        assert req_data == json.dumps({"series": [series]})

        # Enabling compression
        # Metrics
        series = dict(metric="metric.1", points=[(time(), 13.)])
        compressed_series = zlib.compress(json.dumps({"series": [series]}).encode("utf-8"))
        Metric.send(compress_payload=True, attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" in headers
        assert headers["Content-Encoding"] == "deflate"
        assert req_data == compressed_series
        # Distributions
        series = dict(metric='metric.1', points=[(time(), 13.)])
        compressed_series = zlib.compress(json.dumps({"series": [series]}).encode("utf-8"))
        Distribution.send(compress_payload=True, attach_host_name=False, **series)
        _, kwargs = self.request_mock.call_args()
        req_data = kwargs["data"]
        headers = kwargs["headers"]
        assert "Content-Encoding" in headers
        assert headers["Content-Encoding"] == "deflate"
        assert req_data == compressed_series


class TestServiceCheckResource(DatadogAPIWithInitialization):

    def test_service_check_supports_none_parameters(self):
        """
        ServiceCheck should support none parameters

        ```
        $ dog service_check check check_pg host0 1
        ```

        resulted in `RuntimeError: dictionary changed size during iteration`
        """
        ServiceCheck.check(
            check='check_pg', host_name='host0', status=1, message=None,
            timestamp=None, tags=None)


class TestUserResource(DatadogAPIWithInitialization):

    def test_create_user(self):
        User.create(handle="handle", name="name", access_role="ro")
        self.request_called_with(
            "POST", "https://example.com/api/v1/user", data={"handle": "handle", "name": "name", "access_role": "ro"}
        )

    def test_get_user(self):
        User.get("handle")
        self.request_called_with("GET", "https://example.com/api/v1/user/handle")

    def test_update_user(self):
        User.update("handle", name="name", access_role="ro", email="email", disabled="disabled")
        self.request_called_with(
            "PUT",
            "https://example.com/api/v1/user/handle",
            data={"name": "name", "access_role": "ro", "email": "email", "disabled": "disabled"}
        )

    def test_delete_user(self):
        User.delete("handle")
        self.request_called_with("DELETE", "https://example.com/api/v1/user/handle")

    def test_get_all_users(self):
        User.get_all()
        self.request_called_with("GET", "https://example.com/api/v1/user")
